/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "esotk_dfs.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_dfs  DFS related functions
 *
 * TBD
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @brief    Set the group as RAW or CALIB in a frameset
 * @param    set     the input frameset
 * @return
 *   The function returns @c CPL_ERROR_NONE on success or a CPL error
 *   code otherwise.
 */
/*----------------------------------------------------------------------------*/
cpl_error_code esotk_dfs_set_groups(cpl_frameset *set)
{
    /* Check entries */
    cpl_ensure_code(set != NULL, CPL_ERROR_NULL_INPUT);

    /* Initialize */
    cpl_size nframes = cpl_frameset_get_size(set);

    /* Loop on frames */
    for (cpl_size i = 0; i < nframes; i++) {

        cpl_frame  *cur_frame = cpl_frameset_get_position(set, i);
        const char *tag       = cpl_frame_get_tag(cur_frame);

        if (      !strcmp(tag, ESOTK_RAW                        ) ||
                  !strcmp(tag, ESOTK_RAW_CONFMAP                ) ||
                  !strcmp(tag, ESOTK_RAW_BPM                    ) ||
                  !strcmp(tag, ESOTK_RAW_ERROR                  ) ||
                  !strcmp(tag, ESOTK_RAW_VARIANCE               ) ||
                  !strcmp(tag, ESOTK_SPECTRUM_1D                ) ){

            /* RAW frames */
            cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_RAW);

        } else if (!strcmp(tag, ESOTK_OBJ_MASK                  ) ||
                   !strcmp(tag, ESOTK_MASTER_FRINGE             ) ||
                   !strcmp(tag, ESOTK_MASTER_FRINGE_ERROR       ) ||
                   !strcmp(tag, ESOTK_MASTER_FRINGE_CONTRIBUTION) ||
                   !strcmp(tag, ESOTK_MASTER_FRINGE_BPM         ) ||
                   !strcmp(tag, ESOTK_FRINGE_MASK               ) ||
                   !strcmp(tag, ESOTK_FRINGE_CORRECTED          ) ||
                   !strcmp(tag, ESOTK_FRINGE_CORRECTED_BPM      ) ||
                   !strcmp(tag, ESOTK_FRINGE_CORRECTED_ERROR    ) ||
	           !strcmp(tag, ESOTK_WAVELENGTHS_DEST_COLLAPSE  ) ||
                   !strcmp(tag, ESOTK_EOP_PARAM                 ) ){

            /* CALIB frames */
            cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_CALIB);

        } else {

        	/* unknown-frame */
            cpl_frame_set_group(cur_frame, CPL_FRAME_GROUP_NONE);
            cpl_msg_warning(cpl_func, "Frame:%lld with tag:%s, unknown!", i, tag);
        }
    }

    return cpl_error_get_code();
}

/**@}*/
