/*
 * This file is part of the HDRLDEMO pipeline
 * Copyright (C) 2013 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 *                              Includes
 */
/*----------------------------------------------------------------------------*/

#include <string.h>
#include <math.h>

#include "esotk_idp.h"


/*----------------------------------------------------------------------------*/
/**
 *                              Defines
 */
/*----------------------------------------------------------------------------*/

/* Number of IDP keywords in each list */
#define ESOTK_IDP_PRI_NUM_KEYWORDS         40			/* If you add/remove keyword, this number need to be changed */
#define ESOTK_IDP_EXT_NUM_KEYWORDS         26     /* If you add/remove keyword, this number need to be changed */


/*----------------------------------------------------------------------------*/
/**
 *                              Data structures
 */
/*----------------------------------------------------------------------------*/

typedef enum {
  ESOTK_IDP_CHECK_UNSPECIFIED  = 0,   /* Not apply check */
  ESOTK_IDP_CHECK_CONSISTENT      ,   /* Check and apply consistency           rule of the keyword among the input IDPs */
  ESOTK_IDP_CHECK_PROPAGATE       ,   /* Check and apply propagate             rule of the keyword among the input IDPs */
  ESOTK_IDP_CHECK_PROPAGATE_IGNORE,   /* Check and apply propagate with ignore rule of the keyword among the input IDPs */
  ESOTK_IDP_CHECK_RECALCULATE         /* Apply recalculate rule of the keyword */
} type_check;

typedef struct {
  const char       *name;             /* Property name to check */
  cpl_type         type;              /* Property type to check */
  type_check       check;             /* Type of check: CONSISTENT, PROPAGATE, RECALCULATE */
  cpl_propertylist *plist;            /* Result cpl_propertylist */
} property_check;

typedef enum {
  ESOTK_IDP_OP_UNSPECIFIED = 0,       /* Not apply operation */
  ESOTK_IDP_OP_INT_SUM        ,       /* Calculate int    summation of one keyword in all of input IDP files */
  ESOTK_IDP_OP_INT_MIN        ,       /* Calculate int    minimum   of one keyword in all of input IDP files */
  ESOTK_IDP_OP_INT_MAX        ,       /* Calculate int    maximum   of one keyword in all of input IDP files */
  ESOTK_IDP_OP_DOUBLE_SUM     ,       /* Calculate double summation of one keyword in all of input IDP files */
  ESOTK_IDP_OP_DOUBLE_MIN     ,       /* Calculate double minimum   of one keyword in all of input IDP files */
  ESOTK_IDP_OP_DOUBLE_MAX     ,       /* Calculate double maximum   of one keyword in all of input IDP files */
} type_op;


typedef int (*f_compare)(const void *, const void *);


/*----------------------------------------------------------------------------*/
/**
 *                              Functions prototypes
 */
/*----------------------------------------------------------------------------*/

/* Check that the data in the inputs parameters are correct */
static cpl_error_code esotk_idp_check_inputs(
    cpl_frameset              *raws,
    const cpl_table           *table);

/* Cleanup all the variables in the main function */
static void esotk_idp_delete(
    cpl_frameset              *raws,
    cpl_propertylist          *pEnumKeywords_pri,
    property_check            *prop_pri,
    property_check            *prop_ext);

/* Generate list of primary header keywords */
static cpl_error_code esotk_idp_get_keywords_primary(
    property_check            *prop);

/* Generate list of extension header keywords */
static cpl_error_code esotk_idp_get_keywords_extension(
    property_check            *prop);

/* Fill the arrays of property_check with the input data */
static cpl_error_code esotk_idp_fill_properties(
    cpl_frameset              *raws,
    const cpl_table           *table,
    property_check            *prop_pri,
    property_check            *prop_ext);

/* Fill the cpl_propertylist with the propagate keywords */
static cpl_error_code esotk_idp_fill_copy_keys(
    cpl_frameset              *raws,
    cpl_size                  nKeys,
    char                      **copy_keys,
    cpl_propertylist          *pEnumKeywords_pri);

/* Check and apply the consistency rule to a keyword */
static cpl_error_code esotk_idp_consistent(
    cpl_frameset              *raws,
    const cpl_boolean         isPrimary,
    property_check            *key);

/* Compare if all properties are equal in both cpl_propertylist */
static cpl_boolean esotk_idp_consistent_propertylist_equals(
    cpl_propertylist          *plist1,
    cpl_propertylist          *plist2);

/* Check and apply the propagate rule (only check first IDP frame) */
static cpl_error_code esotk_idp_propagate(
    cpl_frameset              *raws,
    const cpl_boolean         isPrimary,
    property_check            *key);

/* Apply recalculate rule of the keyword */
static cpl_error_code esotk_idp_recalculate(
    cpl_frameset              *raws,
    const cpl_table           *table,
    const cpl_boolean         isPrimary,
    property_check            *key,
    property_check            *keywords);

/* Apply operation over all keyword in the input IDPs that match with the key */
static cpl_error_code esotk_idp_recalculate_apply_operation(
    cpl_frameset              *raws,
    const cpl_boolean         isPrimary,
    property_check            *key,
    type_op                   op,
    int                       *ival,
    double                    *dval);

/* Generate a cpl_propertylist with the unique keywords in all of the IDP input files using the arrays include and exclude */
static cpl_propertylist * esotk_idp_recalculate_get_unique_keywords(
    cpl_frameset             *raws,
    const cpl_boolean         isPrimary,
    cpl_type                  type,
    cpl_size                  nKeywords_include,
    property_check            **keywords_include,
    cpl_size                  nKeywords_exclude,
    property_check            **keywords_exclude,
    const char                *template);

/* Compare two int cpl_property */
static int esotk_idp_recalculate_get_unique_keywords_sort_int(
    const cpl_property        *p1,
    const cpl_property        *p2);

/* Compare two string cpl_property */
static int esotk_idp_recalculate_get_unique_keywords_sort_string(
    const cpl_property        *p1,
    const cpl_property        *p2);

/* Get double value of the keyword name in the array of previously calculated proprety_check */
static cpl_error_code esotk_idp_recalculate_get_double(
    property_check            *keywords,
    cpl_size                  size,
    const char                *name,
    double                    *value);

/* Fill the headers with the appropriate previously calculated property_check array values */
static cpl_error_code esotk_idp_fill_headers_with_idp_properties(
    cpl_propertylist          *pEnumKeywords_pri,
    property_check            *prop_pri,
    property_check            *prop_ext,
    cpl_propertylist          *pri_header,
    cpl_propertylist          *ext_header);

/* Function for sort in ascending order elements of the type double */
static int cmp_double_asc(
    const void                *a,
    const void                *b);


/*----------------------------------------------------------------------------*/
/**
 *                          Static variables
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_idp  IDP related functions
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 *                              Functions code
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @brief   Fill the primary and the extension header with keyword necessary to IDP standard
 *
 * @param frameset                input set of frames
 * @param table                   input cpl_table with the IDP calculated data
 * @param header_pri              in/out cpl_propertylist for the primary header
 * @param header_ext              in/out cpl_propertylist for the extension header
 * @param deleteColumnDefinition  bool that explain if must be delete the column definition
 * @param nKeys                   Number of enumerate keywords
 * @param copy_keys               List of keywords for copy in the output file
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
cpl_error_code esotk_idp_fill_headers(
    cpl_frameset       *frameset,
    const cpl_table    *table,
    cpl_propertylist   *header_pri,
    cpl_propertylist   *header_ext,
    cpl_boolean        deleteColumnDefinition,
    cpl_size           nKeys,
    char               **copy_keys)
{
  /* Check inputs */
  cpl_error_ensure(frameset && table && header_pri && header_ext, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Input NULL");

  /* Select the raw frames */
  cpl_frameset *raws = cpl_frameset_new();
  cpl_size nframes = cpl_frameset_get_size(frameset);
  for (cpl_size i = 0; i < nframes; i++) {
      const cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);
      if (cpl_frame_get_group(cur_frame) == CPL_FRAME_GROUP_RAW) {
          cpl_frameset_insert(raws, cpl_frame_duplicate(cur_frame));
      }
  }

  /* Check content of inputs */
  cpl_error_code err = esotk_idp_check_inputs(raws, table);
  if (err != CPL_ERROR_NONE) {
      esotk_idp_delete(raws, NULL, NULL, NULL);
      return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Problem with the inputs cpl_frameset and/or cpl_table");
  }

  /* List of KEYWORDS in the primary and extension header */
  cpl_propertylist *pCopyKeywords_pri = cpl_propertylist_new();  /* Used to copy keywords in the primary output file                            */
  property_check   prop_pri[ESOTK_IDP_PRI_NUM_KEYWORDS];         /* Used in the loops, number of keyword in primary   header defined in the top */
  property_check   prop_ext[ESOTK_IDP_EXT_NUM_KEYWORDS];         /* Used in the loops, number of keyword in extension header defined in the top */


  /* Generate list of primary header keywords */
  err = esotk_idp_get_keywords_primary(prop_pri);
  if (err != CPL_ERROR_NONE) {
      esotk_idp_delete(raws, pCopyKeywords_pri, prop_pri, prop_ext);
      return err;
  }

  /* Generate list of extension header keywords */
  err = esotk_idp_get_keywords_extension(prop_ext);
  if (err != CPL_ERROR_NONE) {
      esotk_idp_delete(raws, pCopyKeywords_pri, prop_pri, prop_ext);
      return err;
  }

  /* Fill idp_properties */
  err = esotk_idp_fill_properties(raws, table, prop_pri, prop_ext);
  if (err != CPL_ERROR_NONE) {
      esotk_idp_delete(raws, pCopyKeywords_pri, prop_pri, prop_ext);
      return err;
  }

  /* Fill enumerate keywords that need be propagate in the primary header */
  err = esotk_idp_fill_copy_keys(raws, nKeys, copy_keys, pCopyKeywords_pri);
  if (err != CPL_ERROR_NONE) {
      esotk_idp_delete(raws, pCopyKeywords_pri, prop_pri, prop_ext);
      return err;
  }

  /* Fill headers with the filled idp_properties */
  err = esotk_idp_fill_headers_with_idp_properties(pCopyKeywords_pri, prop_pri, prop_ext, header_pri, header_ext);
  if (err != CPL_ERROR_NONE) {
      esotk_idp_delete(raws, pCopyKeywords_pri, prop_pri, prop_ext);
      return err;
  }

  /* Cleanup */
  esotk_idp_delete(raws, pCopyKeywords_pri, prop_pri, prop_ext);

  /* Remove all of keywords in the extension header refered the column definition */
  if (deleteColumnDefinition) {
      cpl_propertylist_erase_regexp(header_ext, "^"ESOTK_IDP_EXT_TTYPE, 0);
      cpl_propertylist_erase_regexp(header_ext, "^"ESOTK_IDP_EXT_TFORM, 0);
      cpl_propertylist_erase_regexp(header_ext, "^"ESOTK_IDP_EXT_TUTYP, 0);
      cpl_propertylist_erase_regexp(header_ext, "^"ESOTK_IDP_EXT_TUNIT, 0);
      cpl_propertylist_erase_regexp(header_ext, "^"ESOTK_IDP_EXT_TUCD,  0);
  }

  return err;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Update the keywords in the headers related with the 'WAVE' units
 *
 * @param header_pri  in/out cpl_propertylist for the primary header
 * @param header_ext  in/out cpl_propertylist for the extension header
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
cpl_error_code esotk_idp_update_wave_unit(
    cpl_propertylist   *header_pri,
    cpl_propertylist   *header_ext)
{
  /* Check inputs */
  cpl_error_ensure(header_pri && header_ext,
                   CPL_ERROR_NULL_INPUT, return CPL_ERROR_NULL_INPUT,
                       "Input NULL");

  /* Check that exist the keyword ESOTK_IDP_EXT_TUNIT1 */
  cpl_error_ensure(cpl_propertylist_has(header_ext, ESOTK_IDP_EXT_TUNIT1),
                   CPL_ERROR_ILLEGAL_INPUT, return CPL_ERROR_ILLEGAL_INPUT,
                       "Doesn't exist keyword[%s] in the extension header", ESOTK_IDP_EXT_TUNIT1);

  /* Check the units of 'WAVE', must be in nm . T o*/
  const char *unit_wave = cpl_propertylist_get_string(header_ext, ESOTK_IDP_EXT_TUNIT1);
  if (strcmp(unit_wave, ESOTK_SPECTRUM1D_TUNIT_NM)) {

      /* Check that exist the keywords */
      cpl_boolean exist_WAVEMIN  = cpl_propertylist_has(header_pri, ESOTK_IDP_PRI_WAVELMIN);
      cpl_boolean exist_WAVEMAX  = cpl_propertylist_has(header_pri, ESOTK_IDP_PRI_WAVELMAX);
      cpl_boolean exist_SPEC_BIN = cpl_propertylist_has(header_pri, ESOTK_IDP_PRI_SPEC_BIN );
      cpl_boolean exist_SPEC_VAL = cpl_propertylist_has(header_ext, ESOTK_IDP_EXT_SPEC_VAL);
      cpl_boolean exist_SPEC_BW  = cpl_propertylist_has(header_ext, ESOTK_IDP_EXT_SPEC_BW );
      cpl_error_ensure(exist_WAVEMIN && exist_WAVEMAX && exist_SPEC_BIN && exist_SPEC_VAL && exist_SPEC_BW,
                       CPL_ERROR_ILLEGAL_INPUT, return CPL_ERROR_ILLEGAL_INPUT,
                           "Don't exist some/all keywords[%s=%d,%s=%d,%s=%d,%s=%d,%s=%d] in the primary/extension header",
                           ESOTK_IDP_PRI_WAVELMIN, exist_WAVEMIN,  ESOTK_IDP_PRI_WAVELMAX, exist_WAVEMAX,
                           ESOTK_IDP_PRI_SPEC_BIN, exist_SPEC_BIN,
                           ESOTK_IDP_EXT_SPEC_VAL, exist_SPEC_VAL, ESOTK_IDP_EXT_SPEC_BW,  exist_SPEC_BW);

      double wavelmin = -1.;
      double wavelmax = -1.;
      double specbin  = -1.;

      /* If not nm, do a conversion */
      if (!strcmp(unit_wave, ESOTK_SPECTRUM1D_TUNIT_ANGSTROM)) {

          /* Get the values and convert Agstroms to nm */
          wavelmin = 0.1 * cpl_propertylist_get_double(header_pri, ESOTK_IDP_PRI_WAVELMIN);
          wavelmax = 0.1 * cpl_propertylist_get_double(header_pri, ESOTK_IDP_PRI_WAVELMAX);
          specbin  = 0.1 * cpl_propertylist_get_double(header_pri, ESOTK_IDP_PRI_SPEC_BIN);

      } else {

          /* There isn't a knowledge type */
          return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT, "The unit in the 'WAVE' column are = %s (Unknown) !", unit_wave);
      }

      double spec_val = (wavelmax + wavelmin) / 2.;
      double spec_bw  =  wavelmax - wavelmin;

      /* Set values */

      cpl_propertylist_update_double(header_pri, ESOTK_IDP_PRI_WAVELMIN, wavelmin);
      cpl_propertylist_update_double(header_pri, ESOTK_IDP_PRI_WAVELMAX, wavelmax);
      cpl_propertylist_update_double(header_pri, ESOTK_IDP_PRI_SPEC_BIN, specbin);

      cpl_propertylist_update_double(header_ext, ESOTK_IDP_EXT_SPEC_VAL, spec_val);
      cpl_propertylist_update_double(header_ext, ESOTK_IDP_EXT_SPEC_BW,  spec_bw);
  }

  return CPL_ERROR_NONE;
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
 * @brief Check that the data in the inputs parameters are correct
 *
 * @param raws        input frames in cpl_frameset. It assume that all of them are raw
 * @param table       input cpl_table with the IDP calculated data
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_check_inputs(
    cpl_frameset              *raws,
    const cpl_table           *table)
{
  /* Check inputs */
  cpl_error_ensure(raws && table, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  /* Check size */
  cpl_size nframes = cpl_frameset_get_size(raws);
  if (nframes == 0) {
      return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "No raw frames in the input frameset");
  }

  /* Loop on frames */
  for (cpl_size i = 0; i < nframes; i++) {

      /* Extract each frame */
      const cpl_frame *cur_frame  = cpl_frameset_get_position(raws, i);
      const char      *filename   = cpl_frame_get_filename(cur_frame);

      /* Check that the number of extension is exactly one */
      cpl_size nExtensions = cpl_fits_count_extensions(filename);
      if (nExtensions != 1) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                       "Input frame[%lld]=%s have %lld extensions --> expected 1",
                                       i, filename, nExtensions);
      }
  }

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Cleanup all the variables in the main function
 *
 * @param raws               local cpl_frameset generate only with the raws cpl_frame's
 * @param pCopyKeywords_pri  List of keywords for copy in the primary   header
 * @param prop_pri           List of keywords          in the primary   header
 * @param prop_ext           List of keywords          in the extension header
 *
 */
/*----------------------------------------------------------------------------*/
static void esotk_idp_delete(
    cpl_frameset              *raws,
    cpl_propertylist          *pCopyKeywords_pri,
    property_check            *prop_pri,
    property_check            *prop_ext)
{
  /* Delete the local raw cpl_frameset */
  if (raws) cpl_frameset_delete(raws);

  /* Delete the extra keyword added in the primary header */
  if (pCopyKeywords_pri) {
      cpl_propertylist_delete(pCopyKeywords_pri);
  }

  /* Delete properties */
  property_check *keywords;
  cpl_size       size;
  for (cpl_size extension = 0; extension <= 1; extension++) {

      /* Get data for primary/extension header */
      if (extension == 0) {
          keywords = prop_pri;
          size     = ESOTK_IDP_PRI_NUM_KEYWORDS;
      } else {
          keywords = prop_ext;
          size     = ESOTK_IDP_EXT_NUM_KEYWORDS;
      }

      /* Delete storage cpl_propertylist in each property_check */
      if (keywords) {
          for (cpl_size i = 0; i < size; i++) {
              if (keywords[i].plist) {
                  cpl_propertylist_delete(keywords[i].plist);
              }
          }
      }
  }
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Generate list of primary header keywords
 *
 * @param prop     input empty array with the space allocate for all of primary header keywords
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_get_keywords_primary(
    property_check            *prop)
{
  /* Check inputs */
  cpl_error_ensure(prop, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null input");

  /* Local variables */
  cpl_size i = -1;

  /* Fill all keywords adding your own rule to apply. Initialize to NULL the internal cpl_propertylist */

  prop[++i] = (property_check){ESOTK_IDP_PRI_SIMPLE,   CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_BITPIX,   CPL_TYPE_INT   , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_NAXIS,    CPL_TYPE_INT   , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_EXTEND,   CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_ORIGIN,   CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_TELESCOP, CPL_TYPE_STRING, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_INSTRUME, CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_DISPELEM, CPL_TYPE_STRING, ESOTK_IDP_CHECK_PROPAGATE_IGNORE, NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_SPECSYS,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_OBJECT,   CPL_TYPE_STRING, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_EXT_OBJ,  CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_PROPAGATE_IGNORE, NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_RA,       CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_DEC,      CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_EQUINOX,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_RADESYS,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_EXPTIME,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_TEXPTIME, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_MJD_OBS,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_MJD_END,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_PROG_ID,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_OBID,     CPL_TYPE_INT,    ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_M_EPOCH,  CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_PROV,     CPL_TYPE_STRING, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_PROCSOFT, CPL_TYPE_STRING, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_OBSTECH,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_PRODCATG, CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_FLUXCAL,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_CONTNORM, CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_WAVELMIN, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_WAVELMAX, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_SPEC_BIN, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_TOT_FLUX, CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_FLUXERR,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE_IGNORE, NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_REFERENC, CPL_TYPE_STRING, ESOTK_IDP_CHECK_PROPAGATE_IGNORE, NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_NCOMBINE, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_SNR,      CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_SPEC_RES, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_PRO_CATG, CPL_TYPE_STRING, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_CHECKSUM, CPL_TYPE_STRING, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_PRI_DATASUM,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_RECALCULATE     , NULL};

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Generate list of extension header keywords
 *
 * @param prop     input empty array with the space allocate for all of extension header keywords
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_get_keywords_extension(
    property_check            *prop)
{
  /* Check inputs */
  cpl_error_ensure(prop, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null input");

  /* Local variables */
  cpl_size i = -1;

  /* Fill all keywords adding your own rule to apply. Initialize to NULL the internal cpl_propertylist */

  prop[++i] = (property_check){ESOTK_IDP_EXT_XTENSION, CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_BITPIX,   CPL_TYPE_INT   , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_NAXIS,    CPL_TYPE_INT   , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_VOCLASS,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_VOPUB,    CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TITLE,    CPL_TYPE_STRING, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_OBJECT,   CPL_TYPE_STRING, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_RA,       CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_DEC,      CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_APERTURE, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_PROPAGATE       , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TELAPSE,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TMID,     CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_SPEC_VAL, CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_SPEC_BW,  CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TFIELDS,  CPL_TYPE_INT   , ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_NELEM,    CPL_TYPE_INT   , ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TTYPE,    CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TUTYP,    CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TUNIT,    CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TUCD,     CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TDMIN1,   CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_TDMAX1,   CPL_TYPE_DOUBLE, ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_EXTNAME,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_INHERIT,  CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_CONSISTENT      , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_CHECKSUM, CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_RECALCULATE     , NULL};
  prop[++i] = (property_check){ESOTK_IDP_EXT_DATASUM,  CPL_TYPE_BOOL  , ESOTK_IDP_CHECK_RECALCULATE     , NULL};

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Fill the arrays of property_check with the input data
 *
 * @param raws        local cpl_frameset generate only with the raws cpl_frame's
 * @param table       input cpl_table with the IDP calculated data
 * @param prop_pri    List of keywords in the primary   header
 * @param prop_ext    List of keywords in the extension header
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_fill_properties(
    cpl_frameset              *raws,
    const cpl_table           *table,
    property_check            *prop_pri,
    property_check            *prop_ext)
{
  /* Check inputs */
  cpl_error_ensure(raws && table && prop_pri && prop_ext, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  property_check *keywords;
  cpl_size       size;
  cpl_boolean    isPrimary;
  cpl_error_code err = CPL_ERROR_NONE;
  for (cpl_size extension = 0; extension <= 1 && !err; extension++) {

      /* Get data for primary/extension header */
      if (extension == 0) {
          keywords  = prop_pri;
          size      = ESOTK_IDP_PRI_NUM_KEYWORDS;
          isPrimary = CPL_TRUE;
      } else {
          keywords  = prop_ext;
          size      = ESOTK_IDP_EXT_NUM_KEYWORDS;
          isPrimary = CPL_FALSE;
      }

      /* Loop over the list of keywords */
      for (cpl_size i = 0; i < size && !err; i++) {
          property_check *key = &(keywords[i]);

          /* Check property rule */
          switch (key->check) {
            case ESOTK_IDP_CHECK_CONSISTENT :
              /* cpl_msg_info(cpl_func, "keyword[%s] -> Check consistency", key->name); */
              err = esotk_idp_consistent( raws,        isPrimary, key);
              break;
            case ESOTK_IDP_CHECK_PROPAGATE :
            case ESOTK_IDP_CHECK_PROPAGATE_IGNORE :
              /* cpl_msg_info(cpl_func, "keyword[%s] -> Propagate value", key->name); */
              err = esotk_idp_propagate(  raws,        isPrimary, key);
              break;
            case ESOTK_IDP_CHECK_RECALCULATE :
              /* cpl_msg_info(cpl_func, "keyword[%s] -> Recalculate value", key->name); */
              err = esotk_idp_recalculate(raws, table, isPrimary, key, prop_pri);
              break;
            default :
              return cpl_error_set_message(cpl_func, CPL_ERROR_UNSPECIFIED, "Non specified error decode type of check keyword");
          }
      }
  }

  return err;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Fill the cpl_propertylist with the propagate keywords
 *
 * @param raws               local cpl_frameset generate only with the raws cpl_frame's
 * @param nKeys              Number of keywords in the list of keywords
 * @param copy_keys          List of keywords that we need add
 * @param pCopyKeywords_pri  List with the propagate keywords
 *
 * @return cpl_propertylist with a list of propagate keywords adding the number of raw file of NULL if there isn't keywords in the list
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_fill_copy_keys(
    cpl_frameset              *raws,
    cpl_size                  nKeys,
    char                      **copy_keys,
    cpl_propertylist          *pCopyKeywords_pri)
{
  /* Check parameters */
  cpl_error_ensure(raws, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  /* Loop on frames */
  cpl_size nframes = cpl_frameset_get_size(raws);
  for (cpl_size num_frame = 0; num_frame < nframes; num_frame++) {

      /* Load primary header */
      const cpl_frame  *frame = cpl_frameset_get_position(raws, num_frame);
      cpl_propertylist *plist = cpl_propertylist_load(cpl_frame_get_filename(frame), 0);

      /* Loop on enumerate keywords */
      for (cpl_size key = 0; key < nKeys; key++) {

          /* Check if the keyword exist in the list */
          if(cpl_propertylist_has(plist, copy_keys[key]) ) {

              /* Create the new property */
              cpl_property *prop       = cpl_property_duplicate(cpl_propertylist_get_property_const(plist, copy_keys[key]));
              const char   *orig_name  = cpl_property_get_name(prop);

              /* New stacking prefix */
              const char   *prefix_key = "ESO STACK ";
              const char   *prefix     = "ESO ";
              int          prefix_len  = strlen(prefix);

              /* Create the new name */
              char *name = NULL;
              if (strncmp(orig_name, prefix, prefix_len) == 0) {

                  int  len     = strlen(orig_name);
                  int  new_len = len - prefix_len;
                  char *substr = cpl_malloc((new_len +1) * sizeof(char));

                  memcpy(substr, &orig_name[prefix_len], new_len);
                  substr[new_len] = '\0';

                  name = cpl_sprintf("%s%s%lld", prefix_key, substr, num_frame + 1);
                  cpl_free(substr);

              } else {

                  name = cpl_sprintf("%s%s%lld", prefix_key, orig_name, num_frame + 1);
              }

              /* Change name and add property inside the list */
              cpl_property_set_name(prop, name);
              cpl_propertylist_append_property(pCopyKeywords_pri, prop);

              /* Cleanup */
              cpl_free(name);
              cpl_property_delete(prop);
          }
      }

      /* Cleanup */
      cpl_propertylist_delete(plist);
 }

  return cpl_error_get_code();
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Check and apply the consistency rule to a keyword
 *
 * @param raws        local cpl_frameset generate only with the raws cpl_frame's
 * @param isPrimary   cpl_boolean that indicate if the keyword belongs to the primary header or not
 * @param key         in/out variable to check. This variable will be modified in it internal cpl_propertylist
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 * Description:
 *
 *      if (!exits keyword in all IDP) { continue; }
 *      else if(!equal keyword in all IDP) { throw error and exit function; }
 *      else { copy value in the output cpl_propertylist of key; }
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_consistent(
    cpl_frameset              *raws,
    const cpl_boolean         isPrimary,
    property_check            *key)
{
  /* Check parameters */
  cpl_error_ensure(raws && key, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  /* Local variables */
  cpl_boolean data_not_found = CPL_FALSE;

  /* Loop on frames */
  cpl_size nframes = cpl_frameset_get_size(raws);
  for (cpl_size i = 0; i < nframes; i++) {

      const cpl_frame *cur_frame = cpl_frameset_get_position(raws, i);
      const char      *filename  = cpl_frame_get_filename(cur_frame);

      /* Get correct header keywords: Primary/Extension */
      cpl_errorstate preState = cpl_errorstate_get();

      /* Generate patron: If it have '*' is a regular expression and we don't put the end symbol */
      char *regexp;
      if ( (key->name)[strlen(key->name) - 1] == '*' ) regexp	= cpl_sprintf("^%s", key->name);
      else regexp	= cpl_sprintf("^%s$", key->name);

      cpl_propertylist *header  = isPrimary ? cpl_propertylist_load_regexp(filename, 0, regexp, 0)
          : cpl_propertylist_load_regexp(filename, 1, regexp, 0);
      cpl_free(regexp);

      if (!header){

          /* Mark as data not found */
          data_not_found = CPL_TRUE;

          cpl_errorstate_set(preState);

          /* Check consistency */
          if (key->plist) {
              return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                           "The keyword[%s] isn't consistent. There isn't in the IDP: %s, but yes in others", key->name, filename);
          }

      } else {

          /* If not found previously */
          if (key->plist == NULL) {

              /* First time: Storage values */
              key->plist = header;

              /* Check consistency between the storage and the new header. It's not allow found only in several, need to be in all or any */
              if (data_not_found) {
                  return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                               "The keyword[%s] isn't consistent. There is in the IDP: %s, but not in others", key->name, filename);
              }

          } else {

              /* Next times: Check consistency */
              if (esotk_idp_consistent_propertylist_equals(key->plist, header) == CPL_FALSE) {

                  cpl_msg_error(cpl_func, "Error not consistency in the keyword[%s]", key->name);

                  /* Show inconsistency */
                  cpl_msg_info(cpl_func, "Error not consistency in the keyword[%s] --> dump key->plist", key->name);
                  cpl_propertylist_dump(key->plist, NULL);
                  cpl_msg_info(cpl_func, "Error not consistency in the keyword[%s] --> dump header", key->name);
                  cpl_propertylist_dump(header, NULL);
                  cpl_propertylist_delete(header);

                  /* Return error consistent */
                  return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                               "The keyword[%s] among IDPs, are not consistent", key->name);
              }

              /* Cleanup */
              cpl_propertylist_delete(header);
          }
      }
  }

  /*cpl_propertylist_dump(key->plist, NULL);
    cpl_msg_info(cpl_func, "keyword[%s] -> Consistent !", key->name);*/

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief  Compare if all properties are equal in both cpl_propertylist
 *
 * @param plist1   cpl_propertylist input to check
 * @param plist2   cpl_propertylist input to check
 *
 * @return CPL_TRUE if yes, CPL_FALSE if not or error
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_boolean esotk_idp_consistent_propertylist_equals(
    cpl_propertylist          *plist1,
    cpl_propertylist          *plist2)
{
  /* Check NULL inputs */
  if ((plist1 && !plist2) || (!plist1 && plist2)) return CPL_FALSE;

  /* Check if the 2 list are empty (in that case equals) */
  if (!plist1 && !plist2) return CPL_TRUE;

  /* Check same quantity of cpl_properties in both cpl_propertylist */
  cpl_size size1 = cpl_propertylist_get_size(plist1);
  cpl_size size2 = cpl_propertylist_get_size(plist2);
  if (size1 != size2) return CPL_FALSE;

  /* Check each cpl_property */
  for (cpl_size i = 0; i < size1; i++) {
      cpl_property *prop = cpl_propertylist_get(plist1, i);

      const char *name = cpl_property_get_name(prop);
      cpl_type   type  = cpl_property_get_type(prop);

      /* Compare types */
      switch (type) {
        case CPL_TYPE_BOOL :
          if (cpl_property_get_bool(prop) != cpl_propertylist_get_bool(plist2, name)) {
              return CPL_FALSE;
          }
          break;
        case CPL_TYPE_INT :
          if (cpl_property_get_int(prop) != cpl_propertylist_get_int(plist2, name)) {
              return CPL_FALSE;
          }
          break;
        case CPL_TYPE_DOUBLE :
          if (cpl_property_get_double(prop) != cpl_propertylist_get_double(plist2, name)) {
              return CPL_FALSE;
          }
          break;
        case CPL_TYPE_STRING :
          if (strcmp(cpl_property_get_string(prop), cpl_propertylist_get_string(plist2, name)) != 0) {
              return CPL_FALSE;
          }
          break;
        default :
          return CPL_FALSE;
      }
  }

  return CPL_TRUE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Check and apply the propagate rule (only check first IDP frame)
 *
 * @param raws        local cpl_frameset generate only with the raws cpl_frame's
 * @param isPrimary   cpl_boolean that indicate if the keyword belongs to the primary header or not
 * @param key         in/out variable to check. This variable will be modified in it internal cpl_propertylist
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 * Description:
 *
 *      if (PROPAGATE_IGNORE key and !found in the first IDP frame) { continue; }
 *      else if(!found in the first IDP frame) { throw error and exit function; }
 *      else { copy value in the output cpl_propertylist of key; }
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_propagate(
    cpl_frameset              *raws,
    const cpl_boolean         isPrimary,
    property_check            *key)
{
  /* Check parameters */
  cpl_error_ensure(raws && key, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  /* Take the first frame */
  const cpl_frame *cur_frame = cpl_frameset_get_position(raws, 0);
  const char      *filename  = cpl_frame_get_filename(cur_frame);

  /* Get correct header keywords: Primary/Extension */
  cpl_errorstate preState = cpl_errorstate_get();

  /* Generate the exact patron */
  char             *regexp = cpl_sprintf("^%s$", key->name);
  cpl_propertylist *header = isPrimary ? cpl_propertylist_load_regexp(filename, 0, regexp, 0)
      : cpl_propertylist_load_regexp(filename, 1, regexp, 0);
  cpl_free(regexp);

  if (!header) {

      /* Set previous errorstate */
      cpl_errorstate_set(preState);

      /* Check if ignore */
      if (key->check != ESOTK_IDP_CHECK_PROPAGATE_IGNORE){
          return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                                       "The keyword[%s] can't be propagate. There isn't data in the first input IDP: %s", key->name, filename);
      }

  } else {

      /* save the header in the output cpl_propertylist of the keyword */
      key->plist = header;
  }

  /*cpl_propertylist_dump(key->plist, NULL);
    cpl_msg_info(cpl_func, "keyword[%s] -> Propagate !", key->name);*/

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Apply recalculate rule of the keyword
 *
 * @param raws        local cpl_frameset generate only with the raws cpl_frame's
 * @param table       input cpl_table with the IDP calculated data
 * @param isPrimary   cpl_boolean that indicate if the keyword belongs to the primary header or not
 * @param key         in/out variable to check. This variable will be modified in it internal cpl_propertylist
 * @param keywords    All keywords to check the output values pre-calculated
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_recalculate(
    cpl_frameset              *raws,
    const cpl_table           *table,
    const cpl_boolean         isPrimary,
    property_check            *key,
    property_check            *keywords)
{
  /* Check parameters */
  cpl_error_ensure(raws && table && key, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  /* Check key and create the list */
  cpl_error_ensure(!key->plist, CPL_ERROR_ILLEGAL_INPUT,
                   return CPL_ERROR_ILLEGAL_INPUT, "The result cpl_property_list it's not empty!");
  key->plist = cpl_propertylist_new();

  cpl_errorstate preState = cpl_errorstate_get();

  cpl_error_code err  = CPL_ERROR_NONE;
  int            ival = 0;
  double         dval = 0.;

  if (   !strcmp(key->name, ESOTK_IDP_PRI_EXPTIME )
      || !strcmp(key->name, ESOTK_IDP_PRI_TEXPTIME)
      || !strcmp(key->name, ESOTK_IDP_PRI_MJD_OBS )
      || !strcmp(key->name, ESOTK_IDP_PRI_MJD_END )
      || !strcmp(key->name, ESOTK_IDP_PRI_NCOMBINE) ){

      type_op  op = ESOTK_IDP_OP_UNSPECIFIED;
      cpl_type type;

      if (       !strcmp(key->name, ESOTK_IDP_PRI_NCOMBINE) ){
          op   = ESOTK_IDP_OP_INT_SUM;
          type = CPL_TYPE_INT;
      } else if (!strcmp(key->name, ESOTK_IDP_PRI_EXPTIME ) ||
          !strcmp(key->name, ESOTK_IDP_PRI_TEXPTIME) ){
          op   = ESOTK_IDP_OP_DOUBLE_SUM;
          type = CPL_TYPE_DOUBLE;
      } else if (!strcmp(key->name, ESOTK_IDP_PRI_MJD_OBS ) ){
          op   = ESOTK_IDP_OP_DOUBLE_MIN;
          type = CPL_TYPE_DOUBLE;
      } else if (!strcmp(key->name, ESOTK_IDP_PRI_MJD_END ) ){
          op   = ESOTK_IDP_OP_DOUBLE_MAX;
          type = CPL_TYPE_DOUBLE;
      } else {
          return cpl_error_set_message(cpl_func, CPL_ERROR_UNSPECIFIED, "Unspecified error");
      }

      /* Generic call to apply one operation about all the IDP files */
      err = esotk_idp_recalculate_apply_operation(raws, isPrimary, key, op, &ival, &dval);
      if (err == CPL_ERROR_NONE) {
          switch(type) {
            case CPL_TYPE_INT    : cpl_propertylist_update_int(   key->plist, key->name, ival); break;
            case CPL_TYPE_DOUBLE : cpl_propertylist_update_double(key->plist, key->name, dval); break;
            default :
              return cpl_error_set_message(cpl_func, CPL_ERROR_UNSPECIFIED, "Unspecified error");
          }
      }

  } else if (   !strcmp(key->name, ESOTK_IDP_PRI_PROG_ID)
             || !strcmp(key->name, ESOTK_IDP_PRI_OBID   )
             || !strcmp(key->name, ESOTK_IDP_PRI_PROV   ) ){

      /* Get unique list of keywords, you can to provide a list with include keywords and exclude keywords */
      cpl_propertylist *unique;
      if (!strcmp(key->name, ESOTK_IDP_PRI_PROG_ID)) {
          property_check *keywords_include[2] = {key, &(property_check){ESOTK_IDP_PRI_PROGID,         CPL_TYPE_STRING, ESOTK_IDP_CHECK_UNSPECIFIED, NULL}};
          property_check *keywords_exclude[1] = {     &(property_check){ESOTK_IDP_PRI_PROG_ID_MULTI,  CPL_TYPE_STRING, ESOTK_IDP_CHECK_UNSPECIFIED, NULL}};
          unique = esotk_idp_recalculate_get_unique_keywords(raws, isPrimary, CPL_TYPE_STRING, 2, keywords_include, 1, keywords_exclude, ESOTK_IDP_PRI_PROGID);
      } else {
          property_check *keywords_include[1] = {key};
          unique = esotk_idp_recalculate_get_unique_keywords(raws, isPrimary, key->type,       1, keywords_include, 0, NULL,             key->name);
      }
      if (!unique) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], not data unique found in the table", key->name);
      }

      /* Get size of the unique cpl_propertylist */
      cpl_size size = cpl_propertylist_get_size(unique);

      /* Check the alternatives */
      if (!strcmp(key->name, ESOTK_IDP_PRI_PROG_ID)) {

          /* In ESOTK_IDP_PRI_PROG_ID --> Adapt PROG_ID and fill ESOTK_IDP_PRI_PROGID-ith */

          /* Fill the unique PROG_ID in the file */
          if (size == 1) {
              cpl_property *prop = cpl_property_duplicate(cpl_propertylist_get(unique, 0));
              cpl_property_set_name(prop, key->name);
              cpl_propertylist_append_property(key->plist, prop);
              cpl_property_delete(prop);
          } else {
              cpl_propertylist_update_string(key->plist, key->name, ESOTK_IDP_PRI_PROG_ID_MULTI);
          }

          /*** Changed the specifications --> Now NCOMBINED is calculated as a SUM of NCOMBINE_inputs ***/
          /* In ESOTK_IDP_PRI_PROV -----> Fill also NCOMBINED */
          /*} else if (!strcmp(key->name, ESOTK_IDP_PRI_PROV) ){

			cpl_propertylist_update_int(key->plist, ESOTK_IDP_PRI_NCOMBINE, (int)size);*/
      }

      /* Add unique list to the output cpl_propertylist */
      if (!( !strcmp(key->name, ESOTK_IDP_PRI_PROG_ID) && size == 1)) {
          cpl_propertylist_append(key->plist, unique);
      }

      /* Cleanup */
      cpl_propertylist_delete(unique);

  } else if (!strcmp(key->name, ESOTK_IDP_PRI_M_EPOCH) ){

      /* Set directly CPL_FALSE value */
      cpl_propertylist_update_bool(key->plist, key->name, CPL_FALSE);

  } else if (!strcmp(key->name, ESOTK_IDP_PRI_PROCSOFT) ){

      /* Set directly the version of the recipe */
      char *version = cpl_strdup(PACKAGE "/" PACKAGE_VERSION);
      cpl_propertylist_update_string(key->plist, key->name, version);
      cpl_free(version);

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_TDMIN1  ) ){

      /* Check if exist the 'WAVE' column */
      if (!cpl_table_has_column(table, "WAVE")) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[\"WAVE\"] not founded in the table", key->name);
      } else {

          /* Calculate the minimum wavelength (IDP format --> array) */
          const cpl_array *array = cpl_table_get_array(table, "WAVE", 0);
          dval = cpl_array_get_min(array);
          cpl_propertylist_update_double(key->plist, key->name, dval);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_TDMAX1  ) ){

      /* Check if exist the 'WAVE' column */
      if (!cpl_table_has_column(table, "WAVE")) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[\"WAVE\"] not founded in the table", key->name);
      } else {

          /* Calculate the maximum wavelength (IDP format --> array) */
          const cpl_array *array = cpl_table_get_array(table, "WAVE", 0);
          dval = cpl_array_get_max(array);
          cpl_propertylist_update_double(key->plist, key->name, dval);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_PRI_WAVELMIN)){

/*
       Check if exist the 'WAVE' column
      if (!cpl_table_has_column(table, "WAVE")) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[\"WAVE\"] not founded in the table", key->name);
      } else {

           Calculate the minimum wavelength (IDP format --> array)
          const cpl_array *array = cpl_table_get_array(table, "WAVE", 0);
          dval = cpl_array_get_min(array);
          cpl_propertylist_update_double(key->plist, key->name, dval);
      }
*/

      /* Get the WAVE and BPM name columns (Can be the normal or the reduced) */
      char *column_wave = NULL;
      char *column_bpm = NULL;
      if (       cpl_table_has_column(table, "WAVE") && cpl_table_has_column(table, ESOTK_IDP_PRI_BPM)) {
          column_wave = cpl_sprintf("WAVE");
          column_bpm = cpl_sprintf(ESOTK_IDP_PRI_BPM);
      } else if (cpl_table_has_column(table, "WAVE") && cpl_table_has_column(table, ESOTK_IDP_PRI_BPM_REDUCED)) {
          column_wave = cpl_sprintf("WAVE");
          column_bpm = cpl_sprintf(ESOTK_IDP_PRI_BPM_REDUCED);
      } else {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[%s,%s] or column[%s,%s] not founded in the table",	key->name,
                                       "WAVE", ESOTK_IDP_PRI_BPM, "WAVE", ESOTK_IDP_PRI_BPM_REDUCED);
      }

      /* Get data arrays from WAVE and BPM */
      cpl_array       *wave = cpl_array_duplicate(cpl_table_get_array(table, column_wave, 0));
      const cpl_array *bpm = cpl_table_get_array(table, column_bpm, 0);
      cpl_free(column_wave);
      cpl_free(column_bpm);

      /* Check WAVE data */
      int null;
      cpl_size size = cpl_array_get_size(wave);
      for (cpl_size i = 0; i < size; i++) {

          /* Check the correct data with the BPM */
          if (cpl_array_get_int(bpm, i, &null) != 0) {
              cpl_array_set_invalid(wave, i);
          }

          /* Check if the data are not NaN */
          if (!isfinite(cpl_array_get_double(wave, i, &null))) {
              cpl_array_set_invalid(wave, i);
          }
      }

      /* Take the minimum of the valid data and update the cpl_property */
      dval = cpl_array_get_min(wave);
      cpl_propertylist_update_double(key->plist, key->name, dval);

      /* Cleanup */
      cpl_array_delete(wave);

  } else if (!strcmp(key->name, ESOTK_IDP_PRI_WAVELMAX)){

/*
       Check if exist the 'WAVE' column
      if (!cpl_table_has_column(table, "WAVE")) {
	  return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
				       "keyword[%s], column[\"WAVE\"] not founded in the table", key->name);
      } else {

	   Calculate the maximum wavelength (IDP format --> array)
	  const cpl_array *array = cpl_table_get_array(table, "WAVE", 0);
	  dval = cpl_array_get_max(array);
	  cpl_propertylist_update_double(key->plist, key->name, dval);
      }
*/

      /* Get the WAVE and BPM name columns (Can be the normal or the reduced) */
      char *column_wave = NULL;
      char *column_bpm = NULL;
      if (       cpl_table_has_column(table, "WAVE") && cpl_table_has_column(table, ESOTK_IDP_PRI_BPM)) {
          column_wave = cpl_sprintf("WAVE");
          column_bpm = cpl_sprintf(ESOTK_IDP_PRI_BPM);
      } else if (cpl_table_has_column(table, "WAVE") && cpl_table_has_column(table, ESOTK_IDP_PRI_BPM_REDUCED)) {
          column_wave = cpl_sprintf("WAVE");
          column_bpm = cpl_sprintf(ESOTK_IDP_PRI_BPM_REDUCED);
      } else {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[%s,%s] or column[%s,%s] not founded in the table",	key->name,
                                       "WAVE", ESOTK_IDP_PRI_BPM, "WAVE", ESOTK_IDP_PRI_BPM_REDUCED);
      }

      /* Get data arrays from WAVE and BPM */
      cpl_array       *wave = cpl_array_duplicate(cpl_table_get_array(table, column_wave, 0));
      const cpl_array *bpm = cpl_table_get_array(table, column_bpm, 0);
      cpl_free(column_wave);
      cpl_free(column_bpm);

      /* Check WAVE data */
      int null;
      cpl_size size = cpl_array_get_size(wave);
      for (cpl_size i = 0; i < size; i++) {

          /* Check the correct data with the BPM */
          if (cpl_array_get_int(bpm, i, &null) != 0) {
              cpl_array_set_invalid(wave, i);
          }

          /* Check if the data are not NaN */
          if (!isfinite(cpl_array_get_double(wave, i, &null))) {
              cpl_array_set_invalid(wave, i);
          }
      }

      /* Take the maximum of the valid data and update the cpl_property */
      dval = cpl_array_get_max(wave);
      cpl_propertylist_update_double(key->plist, key->name, dval);

      /* Cleanup */
      cpl_array_delete(wave);

  } else if (!strcmp(key->name, ESOTK_IDP_PRI_SPEC_BIN) ){

      /* Check if exist the 'WAVE' column */
      if (!cpl_table_has_column(table, "WAVE")) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[\"WAVE\"] not founded in the table", key->name);
      } else {

          /* Duplicate array for not modify anything */
          cpl_array *array = cpl_array_duplicate(cpl_table_get_array(table, "WAVE", 0));
          cpl_size size = cpl_array_get_size(array);

          /* Order the wavelengths because can be unordered */
          double *dbl_array = cpl_array_get_data_double(array);
          f_compare func = &cmp_double_asc;
          qsort(dbl_array, size, sizeof(double), func);

          /* Create the delta array of input wavelegths */
          int null;
          cpl_array *delta = cpl_array_new(size - 1, CPL_TYPE_DOUBLE);
          for (cpl_size i = 0; i < size - 1; i++) {
              dval = cpl_array_get(array, i + 1, &null) - cpl_array_get(array, i, &null);
              cpl_array_set_double(delta, i, dval);
          }

          /* Calculate the mediam of delta wavelength (IDP format --> array) */
          dval = cpl_array_get_median(delta);
          cpl_propertylist_update_double(key->plist, key->name, dval);

          /* Cleanup the temporary array */
          cpl_array_delete(array);
          cpl_array_delete(delta);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_PRI_SNR) ){

      /* Get the SNR and BPM name columns (Can be the normal or the reduced) */
      char *column_snr = NULL;
      char *column_bpm = NULL;
      if (       cpl_table_has_column(table, ESOTK_IDP_PRI_SNR) && cpl_table_has_column(table, ESOTK_IDP_PRI_BPM)) {
          column_snr = cpl_sprintf(ESOTK_IDP_PRI_SNR);
          column_bpm = cpl_sprintf(ESOTK_IDP_PRI_BPM);
      } else if (cpl_table_has_column(table, ESOTK_IDP_PRI_SNR_REDUCED) && cpl_table_has_column(table, ESOTK_IDP_PRI_BPM_REDUCED)) {
          column_snr = cpl_sprintf(ESOTK_IDP_PRI_SNR_REDUCED);
          column_bpm = cpl_sprintf(ESOTK_IDP_PRI_BPM_REDUCED);
      } else {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[%s,%s] or column[%s,%s] not founded in the table",	key->name,
                                       ESOTK_IDP_PRI_SNR, ESOTK_IDP_PRI_BPM, ESOTK_IDP_PRI_SNR_REDUCED, ESOTK_IDP_PRI_BPM_REDUCED);
      }

      /* Get data arrays from SNR and BPM */
      cpl_array       *snr = cpl_array_duplicate(cpl_table_get_array(table, column_snr, 0));
      const cpl_array *bpm = cpl_table_get_array(table, column_bpm, 0);
      cpl_free(column_snr);
      cpl_free(column_bpm);

      /* Check SNR data */
      int null;
      cpl_size size = cpl_array_get_size(snr);
      for (cpl_size i = 0; i < size; i++) {

          /* Check the correct data with the BPM */
          if (cpl_array_get_int(bpm, i, &null) != 0) {
              cpl_array_set_invalid(snr, i);
          }

          /* Check if the data are not NaN */
          if (!isfinite(cpl_array_get_double(snr, i, &null))) {
              cpl_array_set_invalid(snr, i);
          }
      }

      /* Take the median of the valid data and update the cpl_property */
      dval = cpl_array_get_median(snr);
      cpl_propertylist_update_double(key->plist, key->name, dval);

      /* Cleanup */
      cpl_array_delete(snr);

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_TITLE) ){

	  /* Implementing ticket PIPE-11857 */
	  const cpl_frame *cur_frame = cpl_frameset_get_position(raws, 0);
	  const char      *filename  = cpl_frame_get_filename(cur_frame);
	  cpl_propertylist *pri_header =  cpl_propertylist_load(filename, 0);

	  if (cpl_propertylist_has(pri_header, "OBJECT") && cpl_propertylist_has(pri_header, "DATE-OBS"))
	  {
		  char* title;
		  title = cpl_sprintf("%s_%s_STACK",
				  cpl_propertylist_get_string(pri_header, "OBJECT"),
				  cpl_propertylist_get_string(pri_header, "DATE-OBS"));
		  /* Replace spaces with underscores */
		  for (int i = 0; title[i] != '\0'; i++) {
			  if (title[i] == ' ') {
				  title[i] = '_';
			  }
		  }
		  cpl_propertylist_update_string(key->plist, key->name, title);
		  cpl_free(title);
	  } else {
		  /* Set the value directly to "DUMMY" */
		  cpl_propertylist_update_string(key->plist, key->name, "DUMMY");
	  }
	  cpl_propertylist_delete(pri_header);

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_TELAPSE) ){

      /* Get the keywords calculated previously: MJD-END and MJD-OBS */
      double mjd_end;
      err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_MJD_END, &mjd_end);
      if (!err) {
          double mjd_obs;
          err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_MJD_OBS, &mjd_obs);
          if (!err) {

              /* Calculate and set the value */
              dval = (mjd_end - mjd_obs) * 24. * 3600.;
              cpl_propertylist_update_double(key->plist, key->name, dval);
          }
      }

      if (err) {
          cpl_errorstate_set(preState);
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "It can't found %s and %s, maybe not previously calculated", ESOTK_IDP_PRI_MJD_END, ESOTK_IDP_PRI_MJD_OBS);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_TMID) ){

      /* Get the keywords calculated previously: MJD-END and MJD-OBS */
      double mjd_obs;
      err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_MJD_OBS, &mjd_obs);
      if (!err) {
          double mjd_end;
          err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_MJD_END, &mjd_end);
          if (!err) {

              /* Calculate and set the value */
              dval = (mjd_end + mjd_obs) / 2.;
              cpl_propertylist_update_double(key->plist, key->name, dval);
          }
      }

      if (err) {
          cpl_errorstate_set(preState);
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "It can't found %s and %s, maybe not previously calculated", ESOTK_IDP_PRI_MJD_OBS, ESOTK_IDP_PRI_MJD_END);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_SPEC_VAL) ){

      /* Get the keywords calculated previously: WAVELMAX and WAVELMIN */
      double wavelmax;
      err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_WAVELMAX, &wavelmax);
      if (!err) {
          double wavelmin;
          err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_WAVELMIN, &wavelmin);
          if (!err) {

              /* Calculate and set the value */
              dval = (wavelmax + wavelmin) / 2;
              cpl_propertylist_update_double(key->plist, key->name, dval);
          }
      }

      if (err) {
          cpl_errorstate_set(preState);
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "It can't found %s and %s, maybe not previously calculated", ESOTK_IDP_PRI_WAVELMAX, ESOTK_IDP_PRI_WAVELMIN);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_SPEC_BW) ){

      /* Get the keywords calculated previously: WAVELMAX and WAVELMIN */
      double wavelmax;
      err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_WAVELMAX, &wavelmax);
      if (!err) {
          double wavelmin;
          err = esotk_idp_recalculate_get_double(keywords, ESOTK_IDP_PRI_NUM_KEYWORDS, ESOTK_IDP_PRI_WAVELMIN, &wavelmin);
          if (!err) {

              /* Calculate and set the value */
              dval = wavelmax - wavelmin;
              cpl_propertylist_update_double(key->plist, key->name, dval);
          }
      }

      if (err) {
          cpl_errorstate_set(preState);
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "It can't found %s and %s, maybe not previously calculated", ESOTK_IDP_PRI_WAVELMAX, ESOTK_IDP_PRI_WAVELMIN);
      }

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_TFIELDS) ){

      /* Calculate and set the value */
      ival = cpl_table_get_ncol(table);
      cpl_propertylist_update_int(key->plist, key->name, ival);

  } else if (!strcmp(key->name, ESOTK_IDP_EXT_NELEM) ){

      /* Check if exist the 'WAVE' column */
      if (!cpl_table_has_column(table, "WAVE")) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "keyword[%s], column[\"WAVE\"] not founded in the table to get the number of elements", key->name);
      } else {

          /* Calculate and set the value */
          const cpl_array *array = cpl_table_get_array(table, "WAVE", 0);
          ival = cpl_array_get_size(array);
          cpl_propertylist_update_int(key->plist, key->name, ival);
      }

  } else if (   !strcmp(key->name, ESOTK_IDP_PRI_CHECKSUM)
             || !strcmp(key->name, ESOTK_IDP_PRI_DATASUM )
             || !strcmp(key->name, ESOTK_IDP_EXT_CHECKSUM)
             || !strcmp(key->name, ESOTK_IDP_EXT_DATASUM ) ){

      cpl_propertylist_update_string(key->plist, key->name, "");

  } else {

      err = cpl_error_set_message(cpl_func, CPL_ERROR_UNSPECIFIED, "Keyword[%s] not in the recalculation list", key->name);
  }

  /*cpl_propertylist_dump(key->plist, NULL);
    cpl_msg_info(cpl_func, "keyword[%s] -> Recalculated !", key->name);*/

  return err;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Apply operation over all keyword in the input IDPs that match with the key
 *
 * @param raws        local cpl_frameset generate only with the raws cpl_frame's
 * @param isPrimary   cpl_boolean that indicate if the keyword belongs to the primary header or not
 * @param key         in/out variable to check. This variable will be modified in it internal cpl_propertylist
 * @param op          Operation to apply in order to calculate one keyword in all of input IDP files
 * @param ival        Output int    value result to apply the operation op
 * @param dval        Output double value result to apply the operation op
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_recalculate_apply_operation(
    cpl_frameset              *raws,
    const cpl_boolean         isPrimary,
    property_check            *key,
    type_op                   op,
    int                       *ival,
    double                    *dval)
{
  /* Check inputs */
  cpl_error_ensure(raws && key, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null inputs");

  /* Select initial value depending of the operation */

  cpl_type type;
  switch (op) {
    case ESOTK_IDP_OP_INT_SUM    : type = CPL_TYPE_INT;    *ival = 0;       break;
    case ESOTK_IDP_OP_INT_MIN    : type = CPL_TYPE_INT;    *ival = INT_MAX; break;
    case ESOTK_IDP_OP_INT_MAX    : type = CPL_TYPE_INT;    *ival = INT_MIN; break;
    case ESOTK_IDP_OP_DOUBLE_SUM : type = CPL_TYPE_DOUBLE; *dval = 0.;      break;
    case ESOTK_IDP_OP_DOUBLE_MIN : type = CPL_TYPE_DOUBLE; *dval = DBL_MAX; break;
    case ESOTK_IDP_OP_DOUBLE_MAX : type = CPL_TYPE_DOUBLE; *dval = DBL_MIN; break;
    default :
      return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Not recognized operation");
  }

  /* Loop on frames */
  int    current_ival = *ival;
  double current_dval = *dval;
  cpl_size nframes = cpl_frameset_get_size(raws);
  for (cpl_size i = 0; i < nframes; i++) {

      const cpl_frame *cur_frame = cpl_frameset_get_position(raws, i);
      const char      *filename  = cpl_frame_get_filename(cur_frame);

      /* Get correct header keywords: Primary/Extension */
      cpl_errorstate preState = cpl_errorstate_get();

      /* Generate the exact patron */
      char             *regexp = cpl_sprintf("^%s$", key->name);
      cpl_propertylist *header = isPrimary ? cpl_propertylist_load_regexp(filename, 0, regexp, 0)
          : cpl_propertylist_load_regexp(filename, 1, regexp, 0);
      cpl_free(regexp);

      cpl_size size = 0;
      if (header) size = cpl_propertylist_get_size(header);

      if (!header || size == 0) {

          /* Set previous errorstate */
          cpl_errorstate_set(preState);

          /* If header exist cleanup */
          if (header) cpl_propertylist_delete(header);

          /* Throw and error because there isn't data */
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "Not founded keyword[%s] in IDP: %s", key->name, filename);
      }

      switch(type) {
        case CPL_TYPE_INT :    current_ival = cpl_propertylist_get_int(   header, key->name); break;
        case CPL_TYPE_DOUBLE : current_dval = cpl_propertylist_get_double(header, key->name); break;
        default :
          return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Illegal type in the input");
      }

      cpl_propertylist_delete(header);

      switch (op) {
        case ESOTK_IDP_OP_INT_SUM    : *ival += current_ival;                 break;
        case ESOTK_IDP_OP_INT_MIN    : *ival  = CPL_MIN(*ival, current_ival); break;
        case ESOTK_IDP_OP_INT_MAX    : *ival  = CPL_MAX(*ival, current_ival); break;
        case ESOTK_IDP_OP_DOUBLE_SUM : *dval += current_dval;                 break;
        case ESOTK_IDP_OP_DOUBLE_MIN : *dval  = CPL_MIN(*dval, current_dval); break;
        case ESOTK_IDP_OP_DOUBLE_MAX : *dval  = CPL_MAX(*dval, current_dval); break;
        default :
          return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT, "Not recognized operation");
      }
  }

  return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Generate a cpl_propertylist with the unique keywords in all of the IDP input files using the arrays include and exclude
 *
 * @param raws               local cpl_frameset generate only with the raws cpl_frame's
 * @param isPrimary          cpl_boolean that indicate if the keyword belongs to the primary header or not
 * @param type               cpl_type to compare
 * @param nKeywords_include  number of elements in the array keywords_include
 * @param keywords_include   keywords to include in the output cpl_propertylist.
 * @param nKeywords_exclude  number of elements in the array keywords_include
 * @param keywords_exclude   keywords to exclude in the output cpl_propertylist.
 * @param template           String template to put in the name of the output patron in the unique keywords
 *
 * @return cpl_propertylist with the unique keyword list
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_propertylist * esotk_idp_recalculate_get_unique_keywords(
    cpl_frameset             *raws,
    const cpl_boolean         isPrimary,
    cpl_type                  type,
    cpl_size                  nKeywords_include,
    property_check            **keywords_include,
    cpl_size                  nKeywords_exclude,
    property_check            **keywords_exclude,
    const char                *template)
{
  /* Check inputs */
  cpl_error_ensure(raws && keywords_include && (nKeywords_exclude == 0 || keywords_exclude) && template, CPL_ERROR_NULL_INPUT,
                   return NULL, "Null inputs");

  /* Check input type */
  cpl_error_ensure(type == CPL_TYPE_INT || type == CPL_TYPE_STRING, CPL_ERROR_INCOMPATIBLE_INPUT,
                   return NULL, "Incompatible type");

  /* Loop over the keywords in include array */
  cpl_propertylist *plist = cpl_propertylist_new();
  for (cpl_size inc = 0; inc < nKeywords_include; inc++) {
      property_check *key = keywords_include[inc];

      /* Loop on frames */
      cpl_size nframes = cpl_frameset_get_size(raws);
      for (cpl_size i = 0; i < nframes; i++) {

          const cpl_frame *cur_frame = cpl_frameset_get_position(raws, i);
          const char      *filename  = cpl_frame_get_filename(cur_frame);

          /* Get correct header keywords: Primary/Extension */
          cpl_errorstate preState = cpl_errorstate_get();

          /* Generate the regular expression patron: Only need to start with the initial name. It allow to change the end numbers */
          char             *regexp = cpl_sprintf("^%s", key->name);
          cpl_propertylist *header = isPrimary ? cpl_propertylist_load_regexp(filename, 0, regexp, 0)
              : cpl_propertylist_load_regexp(filename, 1, regexp, 0);
          cpl_free(regexp);

          if (!header) {

              /* Set previous errorstate */
              cpl_errorstate_set(preState);

              /* Cleanup and return */
              cpl_propertylist_delete(plist);
              return NULL;

          } else {

              /* Add to the list */
              cpl_propertylist_append(plist, header);
              cpl_propertylist_delete(header);
          }
      }
  }

  /* Sort cpl_propertylist by value of keyword */
  if (type == CPL_TYPE_INT) {
      cpl_propertylist_sort(plist, esotk_idp_recalculate_get_unique_keywords_sort_int   );
  } else {
      cpl_propertylist_sort(plist, esotk_idp_recalculate_get_unique_keywords_sort_string);
  }

  /* Create the new unique array */
  cpl_propertylist *unique = cpl_propertylist_new();

  /* Loop over the all keywords */
  cpl_property *pLast  = NULL;
  cpl_size nProperties = cpl_propertylist_get_size(plist);
  cpl_size ind = 1;
  for (cpl_size i = 0; i < nProperties; i++) {
      cpl_property *prop = cpl_propertylist_get(plist, i);

      /* Check if it's and exclude keyword value (only in strings) */
      cpl_boolean exclude = CPL_FALSE;
      if (type == CPL_TYPE_STRING && nKeywords_exclude > 0) {
          const char *value = cpl_property_get_string(prop);
          for (cpl_size exc = 0; exc < nKeywords_exclude; exc++) {
              if (!strcmp(value, keywords_exclude[exc]->name)) {
                  exclude = CPL_TRUE;
              }
          }
      }

      /* If it's and allow keyword: check the value and save if it's necessary */
      if (!exclude) {

          /* Check if equal to the previous one (if i > 0 --> exit previous) */
          cpl_boolean equals = CPL_TRUE;
          if (pLast) {
              if (type == CPL_TYPE_INT) {
                  equals = (cpl_property_get_int(prop) == cpl_property_get_int(pLast));
              } else {
                  equals = !strcmp(cpl_property_get_string(prop),cpl_property_get_string(pLast));
              }
          }

          /* Save unique property with the new name */
          if (!pLast || !equals) {

              /* Storage the unique value in the unique cpl_propertylist */
              char *new_name = cpl_sprintf("%s%lld", template, ind);
              cpl_property_set_name(prop, new_name);
              cpl_propertylist_append_property(unique, prop);
              cpl_free(new_name);

              /* Storage the actual as a property last added to the list and increase the index of unique keywords */
              pLast = prop;
              ind++;
          }
      }
  }

  /* Cleanup */
  cpl_propertylist_delete(plist);

  return unique;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Compare two int cpl_property
 *
 * @param   p1  First  property to compare
 * @param   p2  Second property to compare
 *
 * @return  An integer less than, equal to, or greater than zero if
 *            the name of p1 is found, respectively, to be less than,
 *            to match, or be greater than that of p2.
 */
/*----------------------------------------------------------------------------*/
static int esotk_idp_recalculate_get_unique_keywords_sort_int(
    const cpl_property        *p1,
    const cpl_property        *p2)
{
  int val1 = cpl_property_get_int(p1);
  int val2 = cpl_property_get_int(p2);

  return val1 < val2 ? -1 : (val1 > val2 ? 1 : 0);
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   Compare two string cpl_property
 *
 * @param   p1  First  property to compare
 * @param   p2  Second property to compare
 *
 * @return  An integer less than, equal to, or greater than zero if
 *            the name of p1 is found, respectively, to be less than,
 *            to match, or be greater than that of p2.
 */
/*----------------------------------------------------------------------------*/
static int esotk_idp_recalculate_get_unique_keywords_sort_string(
    const cpl_property        *p1,
    const cpl_property        *p2)
{
  /* Compare the two string properties values */
  return strcmp(cpl_property_get_string(p1), cpl_property_get_string(p2));
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Get double value of the keyword name in the array of previously calculated proprety_check
 *
 * @param keywords    All keywords to check the output values pre-calculated
 * @param size        Size of the keywords array
 * @param name        Name of the cpl_property
 * @param value       Output value in the property
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_recalculate_get_double(
    property_check            *keywords,
    cpl_size                  size,
    const char                *name,
    double                    *value)
{
  /* Loop over the list of keywords */
  cpl_error_code err = CPL_ERROR_DATA_NOT_FOUND;

  /* Get double value */
  for (cpl_size i = 0; i < size && err; i++) {
      cpl_propertylist *plist = keywords[i].plist;
      if (plist) {
          if (cpl_propertylist_has(plist, name)) {
              *value = cpl_propertylist_get_double(plist, name);
              err = CPL_ERROR_NONE;
          }
      }
  }

  return err;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Fill the headers with the appropriate previously calculated property_check array values
 *
 * @param pCopyKeywords_pri  cpl_propertylist with the extra keywords that need add in the primary header
 * @param prop_pri           List of keywords in the primary   header
 * @param prop_ext           List of keywords in the extension header
 * @param header_pri         in/out cpl_propertylist for the primary header
 * @param header_ext         in/out cpl_propertylist for the extension header
 *
 * @return cpl_error_code if error, CPL_ERROR_NOT in other case
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_idp_fill_headers_with_idp_properties(
    cpl_propertylist          *pCopyKeywords_pri,
    property_check            *prop_pri,
    property_check            *prop_ext,
    cpl_propertylist          *header_pri,
    cpl_propertylist          *header_ext)
{
  /* Check inputs */
  cpl_error_ensure(prop_pri && prop_ext && header_pri && header_ext, CPL_ERROR_NULL_INPUT,
                   return CPL_ERROR_NULL_INPUT, "Null input");

  /* Add the copy keywords of the input files in the primary header */
  if (pCopyKeywords_pri) {
      cpl_propertylist_append(header_pri, pCopyKeywords_pri);
  }

  property_check   *keywords;
  cpl_size         size;
  cpl_propertylist *plist;
  for (cpl_size extension = 0; extension <= 1; extension++) {

      /* Select header */
      if (extension == 0) {
          keywords = prop_pri;
          size     = ESOTK_IDP_PRI_NUM_KEYWORDS;
          plist    = header_pri;
      } else {
          keywords = prop_ext;
          size     = ESOTK_IDP_EXT_NUM_KEYWORDS;
          plist    = header_ext;
      }

      /* Check that all of the cpl_propertylist's are filled */
      cpl_boolean err = CPL_FALSE;
      for (cpl_size i = 0; i < size; i++) {

          cpl_size size_plist = 0;
          if(keywords[i].plist) {
              size_plist = cpl_propertylist_get_size(keywords[i].plist);
          }

          if (!keywords[i].plist || size_plist == 0) {

              /* Double check exceptions with output empty cpl_propertylist, the others need to have cpl_propertylist */
              if (   keywords[i].check != ESOTK_IDP_CHECK_CONSISTENT
                  && keywords[i].check != ESOTK_IDP_CHECK_PROPAGATE_IGNORE
                  && size_plist == 0) {

                  /* It's really and error */
                  cpl_msg_error(cpl_func, "The keyword[%s] with check=%d need to have output value to insert in the header!",
                                keywords[i].name, keywords[i].check);

                  /* Set error */
                  if (!err) err = CPL_TRUE;
              }
          }
      }

      /* Check if there was an error */
      if (err) {
          return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                       "Some keywords don't have cpl_property output value to insert in the header!");
      }

      /* Loop over the list of properties to fill the headers */
      for (cpl_size i = 0; i < size; i++) {
          if (keywords[i].plist) {
              cpl_propertylist_append(plist, keywords[i].plist);
          }
      }
  }

  return CPL_ERROR_NONE;
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief    Function for sort in ascending order elements of the type double
 *
 * @param    a         generic pointer to an element of the array (first  in the comparison)
 * @param    b         generic pointer to an element of the array (second in the comparison)
 *
 * @return   result    -1: If 'a' element is less    than 'b'
 *                      1: If 'a' element is greater than 'b'
 *                      0: If 'a' element is equal   than 'b'
 */
/* ---------------------------------------------------------------------------*/
static int cmp_double_asc(const void * a, const void * b)
{
  const double a_d = *(const double *)a;
  const double b_d = *(const double *)b;

  /* Comparation funtion between double to the qsort */
  return a_d < b_d ? -1 : (a_d > b_d ? 1 : 0);
}
