/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                  Includes
 -----------------------------------------------------------------------------*/

#include "esotk_masterfringe_utils.h"
#include "esotk_utils.h"

#include "hdrl.h"

#include <cpl.h>

/*-----------------------------------------------------------------------------
                            Static Prototypes
 -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                            Internal functions
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_utils  recipe related utilities
 *
 * TBD
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/**
 * @brief  Check all of numbers of things in the framesets
 *
 * @param		fs_mask			input  fringe mask frameset
 *
 * @return		mask			output cpl_mask associate or NULL
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_mask * esotk_masterfringe_load_fringe_mask(cpl_frameset *fs_mask)
{
	/* if there is only one fringe mask frame with one extension it applies globally
	 * to all data else fringe_mask is not set and it is loaded per extension later */

	cpl_mask *mask = NULL;

	if (cpl_frameset_get_size(fs_mask) == 1) {

		const cpl_frame *f_mask = cpl_frameset_get_position(fs_mask, 0);
		const char      *fn     = cpl_frame_get_filename(f_mask);

		/* try 0 and 1 extension */
		cpl_errorstate cleanstate = cpl_errorstate_get();
		if (cpl_frame_get_nextensions(f_mask) <= 1) {

			mask = cpl_mask_load(fn, 0, 0);
			cpl_errorstate_set(cleanstate);

			if (mask == NULL) {
				mask = cpl_mask_load(fn, 0, 1);
				cpl_errorstate_set(cleanstate);
			}
		}

		if (mask) {
			cpl_msg_debug(cpl_func, "Number of fringe Bad pixels: %zu", (size_t)cpl_mask_count(mask));
		}
    }

	return mask;
}

/**
 * @brief  Loop over all frames in the image frameset
 *
 * @param		extension		Number of extension
 * @param		hlist_fringe	hdrl_image list set for add hdrl_images
 * @param		clist_obj		cpl_image  list set for add obj masks
 * @param		fs				Image          frameset
 * @param		fs_bpm			Bad pixel mask frameset
 * @param		fs_obj			Object         frameset
 * @param		fs_err  		Error          frameset
 *
 * @return		cpl_error_code
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code esotk_masterfringe_loop_img_frames(cpl_size extension,
	hdrl_imagelist *hlist_fringe, cpl_imagelist *clist_obj,	cpl_frameset *fs,
	cpl_frameset *fs_bpm, cpl_frameset *fs_obj, cpl_frameset *fs_err)
{
    const cpl_size N_img = cpl_frameset_get_size(fs    );
    const cpl_size N_bpm = cpl_frameset_get_size(fs_bpm);
    const cpl_size N_err = cpl_frameset_get_size(fs_err);
    const cpl_size N_obj = cpl_frameset_get_size(fs_obj);

	/* loop over all the frames */
	for (cpl_size frame = 0; frame < N_img; frame++) {

		cpl_frame  *img_frame = cpl_frameset_get_position(fs, frame);
		const char *img_fn    = cpl_frame_get_filename(img_frame);
		cpl_msg_debug(cpl_func, "frame = %s (%d), extension = %d", img_fn, (int)frame, (int)extension);

		/* load image data */
		cpl_image *img = cpl_image_load(img_fn, HDRL_TYPE_DATA, 0, extension);
		if (cpl_error_get_code() != CPL_ERROR_NONE) {
			if (img) cpl_image_delete(img);
			return cpl_error_set_message(cpl_func, cpl_error_get_code(), "Unable to load image.");
		}

		/* load extension specific bad pixel map */
		if (N_bpm > 0 && cpl_error_get_code() == CPL_ERROR_NONE) {

			cpl_frame *bpm_frame = cpl_frameset_get_position(fs_bpm, frame);

			cpl_mask *bpm = cpl_mask_load(cpl_frame_get_filename(bpm_frame), 0, extension);
			if (cpl_error_get_code() != CPL_ERROR_NONE) {
			    cpl_image_delete(img);
				if (bpm) cpl_mask_delete(bpm);
				return cpl_error_set_message(cpl_func, cpl_error_get_code(), "Unable to load bpm frame.");
			}

			if (bpm) {
				cpl_msg_debug(cpl_func, "Number of Bad pixels: %d", (int)cpl_mask_count(bpm));
				cpl_image_reject_from_mask(img, bpm);
				cpl_mask_delete(bpm);
			}
		}

		/* load specific error image */
		cpl_image *err = NULL;
		if (N_err > 0 && cpl_error_get_code() == CPL_ERROR_NONE) {

			cpl_frame *err_frame = cpl_frameset_get_position(fs_err, frame);

			err = cpl_image_load(cpl_frame_get_filename(err_frame), HDRL_TYPE_ERROR, 0, extension);
			if (cpl_error_get_code() != CPL_ERROR_NONE) {
				cpl_image_delete(img);
				if (err) cpl_image_delete(err);
				return cpl_error_set_message(cpl_func, cpl_error_get_code(), "Unable to load error image.");
			}

			if (!strcmp(cpl_frame_get_tag(err_frame), "RAW_VARIANCE")) {

				/* Transform variance into error map and propagate possible bad pixels */
				cpl_image_power(err, 0.5);

				cpl_mask *bpm_img = cpl_image_get_bpm(img);
				cpl_mask *bpm_err = cpl_image_get_bpm(err);

				cpl_mask_or(bpm_img, bpm_err);
			}
		}

		/* create final hdrl_image */
		hdrl_image *himg = hdrl_image_create(img, err);
		cpl_image_delete(img);
		if(err) cpl_image_delete(err);

		/* Add hdrl_image to the imagelist set */
		hdrl_imagelist_set(hlist_fringe, himg, hdrl_imagelist_get_size(hlist_fringe));
		if (cpl_error_get_code() != CPL_ERROR_NONE) {
			return cpl_error_set_message(cpl_func, cpl_error_get_code(), "hdrl_imagelist_set() error.");
		}

		/* load object mask */
		if (N_obj > 0 && cpl_error_get_code() == CPL_ERROR_NONE) {

			cpl_frame *obj_frame = cpl_frameset_get_position(fs_obj, frame);

			cpl_image *obj = cpl_image_load(cpl_frame_get_filename(obj_frame), CPL_TYPE_INT, 0, extension);
			if (cpl_error_get_code() != CPL_ERROR_NONE) {
				if (obj) cpl_image_delete(obj);
				return cpl_error_set_message(cpl_func, cpl_error_get_code(), "Unable to load obj image.");
			}

			if (obj) {
				cpl_imagelist_set(clist_obj, obj, cpl_imagelist_get_size(clist_obj));
				if (cpl_error_get_code() != CPL_ERROR_NONE) {
					cpl_image_delete(obj);
					return cpl_error_set_message(cpl_func, cpl_error_get_code(), "cpl_imagelist_set() error.");
				}
			}
		}
	}

	return CPL_ERROR_NONE;
}

/**
 * @brief  Clean parameters allocated in the recipe
 *
 * @param		fs				Image          frameset
 * @param		fs_cnf			Confidence     frameset
 * @param		fs_bpm			Bad pixel mask frameset
 * @param		fs_obj			Object         frameset
 * @param		fs_err  		Error          frameset
 * @param		fs_mask			Mask           frameset
 * @param		f_data			Data           frame
 * @param		f_err			Error          frame
 * @param		f_bpm			Bad pixel mask frame
 * @param       mask        	Mask of mask   frameset
 * @param		h_Collapse		collapse parameters in the recipe
 * @param		himg_master		hdrl_image master fringe
 * @param		hlist_fringe	hdrl_image list set for add hdrl_images
 * @param		clist_obj		cpl_image  list set for add obj masks
 * @param		master			hdrl_image
 * @param		contrib_map		cpl_image
 * @param		qctable  		QC table
 *
 */
/* ---------------------------------------------------------------------------*/
void esotk_masterfringe_clean_all_recipe_parameters(
	cpl_frameset *fs, cpl_frameset *fs_cnf, cpl_frameset *fs_bpm,
	cpl_frameset *fs_obj, cpl_frameset *fs_err, cpl_frameset *fs_mask,
	cpl_frame *f_data, cpl_frame *f_err, cpl_frame *f_bpm,
	cpl_mask* mask, hdrl_parameter *h_Collapse,
	hdrl_image *himg_master, hdrl_imagelist *hlist_fringe, cpl_imagelist *clist_obj,
	hdrl_image *master, cpl_image  *contrib_map, cpl_table  *qctable)
{
	esotk_clean_recipe_parameters(fs, fs_cnf, fs_bpm, fs_obj, fs_err, fs_mask,
		f_data, f_err, f_bpm, mask, h_Collapse);

	esotk_masterfringe_clean_loop_recipe_parameters(
		himg_master, hlist_fringe, clist_obj, master, contrib_map, qctable);
}

/**
 * @brief  Clean parameters allocated in the recipe
 *
 * @param		himg_master		hdrl_image master fringe
 * @param		hlist_fringe	hdrl_image list set for add hdrl_images
 * @param		clist_obj		cpl_image  list set for add obj masks
 * @param		master			hdrl_image
 * @param		contrib_map		cpl_image
 * @param		qctable  		QC table
 *
 */
/* ---------------------------------------------------------------------------*/
void esotk_masterfringe_clean_loop_recipe_parameters(
	hdrl_image *himg_master, hdrl_imagelist *hlist_fringe, cpl_imagelist *clist_obj,
	hdrl_image *master, cpl_image  *contrib_map, cpl_table  *qctable)
{
	if(himg_master ) hdrl_image_delete(himg_master);

	if(hlist_fringe) hdrl_imagelist_delete(hlist_fringe);
	if(clist_obj   ) cpl_imagelist_delete( clist_obj);

	if(master      ) hdrl_image_delete(master);
	if(contrib_map ) cpl_image_delete(contrib_map);
	if(qctable     ) cpl_table_delete(qctable);
}

/**@}*/
