/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                  Includes
 -----------------------------------------------------------------------------*/

#include "esotk_utils.h"
#include "esotk_dfs.h"

#include "hdrl.h"

#include <cpl.h>

/*-----------------------------------------------------------------------------
                            Static Prototypes
 -----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
                            Internal functions
 -----------------------------------------------------------------------------*/
static cpl_error_code esotk_fringe_get_qcvalues(const cpl_table* qctable,
												cpl_propertylist* qclist)
{
    cpl_ensure_code(qctable, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(qclist,  CPL_ERROR_NULL_INPUT);


    if (cpl_table_has_column(qctable, "Background_level") &&
        cpl_table_has_column(qctable, "Fringe_amplitude") ){

        cpl_size nrow = cpl_table_get_nrow(qctable);
        for (cpl_size i = 0; i < nrow; i++) {

            char *bkg_name;
            char *bkg_comment;
            char *amp_name;
            char *amp_comment;

            if (nrow == 1) {
                bkg_name    = cpl_sprintf("ESO QC BKGLEVEL");
                amp_name    = cpl_sprintf("ESO QC FRINGEAMP");
                amp_comment = cpl_sprintf("Amplitude of fringe pattern (ADU)");
                bkg_comment = cpl_sprintf("Background level (ADU)");

            } else {
                bkg_name    = cpl_sprintf("ESO QC BKGLEVEL%d",             (int)i + 1);
                amp_name    = cpl_sprintf("ESO QC FRINGEAMP%d",            (int)i + 1);
                amp_comment = cpl_sprintf("Fringe amplitude ima %d (ADU)", (int)i + 1);
                bkg_comment = cpl_sprintf("Background level ima %d (ADU)", (int)i + 1);
            }

            double bkg_value = cpl_table_get_double(qctable, "Background_level", i, NULL);
            double amp_value = cpl_table_get_double(qctable, "Fringe_amplitude", i, NULL);

            cpl_propertylist_update_double(qclist, bkg_name, bkg_value  );
            cpl_propertylist_set_comment(  qclist, bkg_name, bkg_comment);

            cpl_propertylist_update_double(qclist, amp_name, amp_value  );
            cpl_propertylist_set_comment(  qclist, amp_name, amp_comment);

            cpl_free(bkg_name   );
            cpl_free(amp_name   );
            cpl_free(bkg_comment);
            cpl_free(amp_comment);
       }
    }

    return cpl_error_get_code();
}


/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_utils  recipe related utilities
 *
 * TBD
 */
/*----------------------------------------------------------------------------*/

/**@{*/


/*----------------------------------------------------------------------------*/
/**
 * @brief   save an image on a FITS file with proper FITS header
 *
 * @param   procatg    FITS product category
 * @param   recipe     recipe name
 * @param   filename   product filename
 * @param   save_type  cpl image type
 * @param   image      actual image to be saved
 * @param   qctable    table with QC values
 * @param   parlist    input recipe parameters
 * @param   frameset   product set of frames
 * @param   type       start_0, start_1 or append, defines extension to write to
 * @param   inherit    frame
 *
 * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
cpl_error_code esotk_save_image( const char              *procatg,
								 const char              *recipe,
								 const char              *filename,
								 const cpl_type          save_type,
								 const cpl_image         *image,
								 const cpl_table         *qctable,
								 const cpl_parameterlist *parlist,
								 cpl_frameset            *frameset,
								 const char              *type,
								 const cpl_frame         *inherit)
{   
    cpl_ensure_code(procatg && recipe && filename && save_type && image &&
                    frameset && type, CPL_ERROR_NULL_INPUT); 

    /* Add a QC parameter  */
    cpl_propertylist * aplist = cpl_propertylist_new();
    cpl_propertylist * qclist = cpl_propertylist_new();

    /* Add the product category and save image */
    cpl_propertylist_update_string(aplist, CPL_DFS_PRO_CATG, procatg);

    /* in case of rescale parameter set to FALSE */
    if (qctable != NULL) {
        esotk_fringe_get_qcvalues(qctable, qclist);
    } else {
        cpl_msg_debug(cpl_func, "'qctable' is NULL");
    }

    /* save image depending of the type */
    if (strcmp(type, "append") == 0) {

        cpl_image_save(image, filename, save_type, qclist, CPL_IO_EXTEND);

    } else if (strcmp(type, "start_0") == 0) {

        cpl_propertylist_update_string(qclist, CPL_DFS_PRO_CATG, procatg);

        cpl_dfs_save_image(frameset, NULL, parlist, frameset, inherit,
                           image, save_type,
                           recipe, qclist, NULL,
                           PACKAGE "/" PACKAGE_VERSION, filename);

    } else if (strcmp(type, "start_1") == 0) {

        cpl_dfs_save_propertylist(frameset, aplist, parlist, frameset,
                                  inherit, recipe, aplist,
                                  NULL, PACKAGE "/" PACKAGE_VERSION,
                                  filename);

        cpl_image_save(image, filename, save_type, qclist, CPL_IO_EXTEND);

    } else {

        cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
            "The saving mode option '%s' is not a valid value.", type);
    }

    cpl_propertylist_delete(aplist);
    cpl_propertylist_delete(qclist);

    return cpl_error_get_code();
}


/* ---------------------------------------------------------------------------*/
/**
 * @brief Take the entry plist and get the data NAXIS1
 *
 * @param   plist    input property list
 *
 * @return  value of ZNAXIS1 or alternatively NAXIS1 in plist
 *
 * ---------------------------------------------------------------------------*/
int esotk_get_naxis1(const cpl_propertylist *plist)
{
    cpl_ensure(plist != NULL, CPL_ERROR_NULL_INPUT, 0);

    if (cpl_propertylist_has(plist, "ZNAXIS1")) {
        return cpl_propertylist_get_int(plist, "ZNAXIS1");
    } else {
        return cpl_propertylist_get_int(plist, "NAXIS1");
    }
}

/* ---------------------------------------------------------------------------*/
/**
 * @brief Take the entry plist and get the data NAXIS2
 *
 * @param   plist    input property list
 *
 * @return  value of ZNAXIS2 or alternatively NAXIS2 in plist
 */
/* ---------------------------------------------------------------------------*/
int esotk_get_naxis2(const cpl_propertylist *plist)
{
    cpl_ensure(plist != NULL, CPL_ERROR_NULL_INPUT, 0);

    if (cpl_propertylist_has(plist, "ZNAXIS2")) {
        return cpl_propertylist_get_int(plist, "ZNAXIS2");
    } else {
        return cpl_propertylist_get_int(plist, "NAXIS2");
    }
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   check the entries in the recipe and classify the frameset with the tags
 *
 * @param   frameset      input set of frames
 *
 * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
cpl_error_code esotk_check_and_set_groups(
  cpl_frameset *frameset)
{
  /* Check size of frameset for to know if the sof file is not empty */
  cpl_size nframes = cpl_frameset_get_size(frameset);
  if(nframes <= 0){

      /* Empty sof file */
      return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                 "There aren't frames in the frameset");
  } else {
      for (cpl_size i = 0; i < nframes; i++) {

          cpl_frame  *frame    = cpl_frameset_get_position(frameset, i);
          const char *filename = cpl_frame_get_filename(frame);

          /* Check if the FITS file exist and have correct data,
           * return 0 if the fits file is valid without extensions */
          if (cpl_fits_count_extensions(filename) < 0){

              return cpl_error_set_message(cpl_func, cpl_error_get_code(),
                         "Problem with the file '%s' (%s --> Code %d)",
                         filename, cpl_error_get_message(), cpl_error_get_code());
          }
      }
  }

  /* Identify the RAW, CONF and CALIB frames in the input frameset */
  if (esotk_dfs_set_groups(frameset)) {

      /* Error classify frames */
      return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                 "Cannot classify RAW and/or CALIB frames");
  } else {

      /* Check classification */
      for (cpl_size i = 0; i < nframes; i++) {

          cpl_frame       *frame = cpl_frameset_get_position(frameset, i);
          const char      *tag   = cpl_frame_get_tag(frame);
          cpl_frame_group group  = cpl_frame_get_group(frame);

          /* The tag is invalid */
          if (group == CPL_FRAME_GROUP_NONE) {
              return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
                         "Frame:%lld with tag:%s is invalid", i, tag);
          }
      }
  }

  return CPL_ERROR_NONE;
}

/**
 * @brief  Clean parameters allocated in the recipe
 *
 * @param		fs			Image          frameset
 * @param		fs_cnf		Confidence     frameset
 * @param		fs_bpm		Bad pixel mask frameset
 * @param		fs_obj		Object         frameset
 * @param		fs_err  	Error          frameset
 * @param		fs_mask		Mask           frameset
 * @param		f_data		Data           frame
 * @param		f_err		Error          frame
 * @param		f_bpm		Bad pixel mask frame
 * @param       mask        Mask of mask   frameset
 * @param		p	        h_parameter in the recipe
 *
 */
/* ---------------------------------------------------------------------------*/
void esotk_clean_recipe_parameters(
	cpl_frameset *fs, cpl_frameset *fs_cnf, cpl_frameset *fs_bpm,
	cpl_frameset *fs_obj, cpl_frameset *fs_err, cpl_frameset *fs_mask,
	cpl_frame *f_data, cpl_frame *f_err, cpl_frame *f_bpm,
	cpl_mask* mask, hdrl_parameter *p)
{
	if(fs     ) cpl_frameset_delete(fs);
	if(fs_cnf ) cpl_frameset_delete(fs_cnf);
	if(fs_bpm ) cpl_frameset_delete(fs_bpm);
	if(fs_obj ) cpl_frameset_delete(fs_obj);
	if(fs_err ) cpl_frameset_delete(fs_err);
	if(fs_mask) cpl_frameset_delete(fs_mask);

	if(f_data ) cpl_frame_delete(f_data);
	if(f_err  ) cpl_frame_delete(f_err);
	if(f_bpm  ) cpl_frame_delete(f_bpm);

	if(mask   ) cpl_mask_delete(mask);

	if(p) hdrl_parameter_delete(p);
}

/**
 * @brief  Check all of numbers of things in the framesets
 *
 * @param		fs			Image          frameset
 * @param		fs_cnf		Confidence     frameset
 * @param		fs_bpm		Bad pixel mask frameset
 * @param		fs_obj		Object         frameset
 * @param		fs_err  	Error          frameset
 * @param		fs_mask		Mask           frameset
 * @param		f_data		Data           frame
 * @param		f_err		Error          frame
 * @param		f_bpm		Bad pixel mask frame
 * @param       mask        Mask of mask   frameset
 * @param		p	        h_parameter in the recipe
 *
 * @return		cpl_error_code
 *
 */
/* ---------------------------------------------------------------------------*/
cpl_error_code esotk_integrity_checks(
	cpl_frameset *fs, cpl_frameset *fs_cnf, cpl_frameset *fs_bpm,
	cpl_frameset *fs_obj, cpl_frameset *fs_err, cpl_frameset *fs_mask,
	cpl_frame *f_data, cpl_frame *f_err, cpl_frame *f_bpm,
	cpl_mask *mask, hdrl_parameter *p)
{
    const cpl_size N_img  = fs      == NULL ? 0 : cpl_frameset_get_size(fs     );
    const cpl_size N_cnf  = fs_cnf  == NULL ? 0 : cpl_frameset_get_size(fs_cnf );
    const cpl_size N_bpm  = fs_bpm  == NULL ? 0 : cpl_frameset_get_size(fs_bpm );
    const cpl_size N_err  = fs_err  == NULL ? 0 : cpl_frameset_get_size(fs_err );
    const cpl_size N_obj  = fs_obj  == NULL ? 0 : cpl_frameset_get_size(fs_obj );
    const cpl_size N_mask = fs_mask == NULL ? 0 : cpl_frameset_get_size(fs_mask);
    const cpl_size N_ext  = cpl_frame_get_nextensions(cpl_frameset_get_position(fs, 0));

    cpl_msg_info(cpl_func, "Number of raw images found:          %"CPL_SIZE_FORMAT, N_img );
    cpl_msg_info(cpl_func, "Number of confidence maps found:     %"CPL_SIZE_FORMAT, N_cnf);
    cpl_msg_info(cpl_func, "Number of bad-pixel masks found:     %"CPL_SIZE_FORMAT, N_bpm );
    cpl_msg_info(cpl_func, "Number of error/variance maps found: %"CPL_SIZE_FORMAT, N_err );
    cpl_msg_info(cpl_func, "Number of object found:              %"CPL_SIZE_FORMAT, N_obj );
    cpl_msg_info(cpl_func, "Number of masks found:               %"CPL_SIZE_FORMAT, N_mask);
    cpl_msg_info(cpl_func, "Number of extensions found:          %"CPL_SIZE_FORMAT, N_ext );

    /* Integrity checks */
    cpl_error_code e_code = CPL_ERROR_NONE;
    char           *e_str = NULL;

    if (N_img < 1) {
        e_code = CPL_ERROR_DATA_NOT_FOUND;
        e_str  = cpl_sprintf("No RAW frames supplied.");
    }

    if (!e_str && N_cnf > 0) {

       if (N_cnf != N_img) {
           e_code = CPL_ERROR_INCOMPATIBLE_INPUT;
           e_str  = cpl_sprintf("If RAW_CONFMAP frames are supplied, their number must match the number of RAW frames.");
       }

       if (!e_str && esotk_checkmatch(fs, fs_cnf) != CPL_ERROR_NONE) {
           e_code = cpl_error_get_code();
           e_str  = cpl_sprintf("RAW/RAW_CONFMAP structure mismatch.");
       }
    }

    if (!e_str && N_bpm > 0) {

       if (N_bpm != N_img) {
           e_code = CPL_ERROR_INCOMPATIBLE_INPUT;
           e_str  = cpl_sprintf("If RAW_BPM frames are supplied, their number must match the number of RAW frames.");
       }

       if (!e_str && esotk_checkmatch(fs, fs_bpm) != CPL_ERROR_NONE) {
           e_code = cpl_error_get_code();
           e_str  = cpl_sprintf("RAW/RAW_BPM structure mismatch.");
       }
    }

    if (!e_str && N_err > 0) {

       if (N_err != N_img) {
           e_code = CPL_ERROR_INCOMPATIBLE_INPUT;
           e_str  = cpl_sprintf("If RAW_ERR or RAW_VARIANCE frames are supplied, their number must match the number of RAW frames.");
       }

       if (!e_str && esotk_checkmatch(fs, fs_err) != CPL_ERROR_NONE) {
           e_code = cpl_error_get_code();
           e_str  = cpl_sprintf("RAW <-> RAW_ERR/RAW_VARIANCE structure mismatch.");
       }
    }

    if (!e_str && N_obj > 0) {

       if (N_obj != N_img) {
           e_code = CPL_ERROR_INCOMPATIBLE_INPUT;
           e_str  = cpl_sprintf("If OBJ_MASK frames are supplied, their number must match the number of RAW frames.");
       }

       const cpl_size N_ext_obj = cpl_frame_get_nextensions(cpl_frameset_get_position(fs_obj, 0));
       if (!e_str && N_ext_obj < 0) {
           e_code = cpl_error_get_code();
           e_str  = cpl_sprintf("Unable to retrieve number of OBJ_MASK extensions.");
       }

       if (!e_str && N_ext_obj != N_ext) {
           e_code = CPL_ERROR_INCOMPATIBLE_INPUT;
           e_str  = cpl_sprintf("Extension number mismatch between RAW and OBJ_MASK frames.");
       }

       if (!e_str && esotk_checkmatch(fs, fs_obj) != CPL_ERROR_NONE) {
           e_code = cpl_error_get_code();
           e_str  = cpl_sprintf("RAW/OBJ_MASK structure mismatch.");
       }
    }

    /* Only zero or one ESOTK_FRINGE_MASK file is allowed in the sof */
    if (!e_str && N_mask > 1) {

        e_code = CPL_ERROR_UNSUPPORTED_MODE;
        e_str  = cpl_sprintf("Multiple FRINGE_MASK frames not supported.");

    } else if (!e_str && N_mask == 1) {

        const cpl_size N_ext_mask = cpl_frame_get_nextensions(cpl_frameset_get_position(fs_mask, 0));
        if (N_ext_mask < 0) {
            e_code = cpl_error_get_code();
            e_str  = cpl_sprintf("Unable to retrieve number of FRINGE_MASK extensions.");
        }

        /* 3rd condition put here because of criteria implicit in -test -> doublecheck. */
        /* if (N_ext_mask != N_ext && N_ext_mask != 0 && N_ext_mask != 1) */

        /* Here we assume that if a fringe mask fits-file is present in the sof
         * and the image is only in the primary header or in the first extenison
         *  this fringe mask has to be applied to all extensions of the raw ima
         */

        if (!e_str && N_ext_mask != N_ext && N_ext_mask != 0 && N_ext_mask != 1) {
            e_code = CPL_ERROR_INCOMPATIBLE_INPUT;
            e_str  = cpl_sprintf("Extension number mismatch between RAW and FRINGE_MASK frames.");
        }
    }

    if (!e_str && N_ext < 0) {
        e_code = cpl_error_get_code();
        e_str  = cpl_sprintf("Unable to load extensions.");
    }

    /* Throw error */
    if (e_str) {
    	esotk_clean_recipe_parameters(fs, fs_cnf, fs_bpm, fs_obj, fs_err,
    			fs_mask, f_data, f_err, f_bpm, mask, p);
    	cpl_error_set_message(cpl_func, e_code, "%s", e_str);
    	cpl_free(e_str);
    	return cpl_error_get_code();
    }

	return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   perform check of matching number of extensions and dimensions;
 *          (assumes the numbers of frames of 2 input framesets already match)
 *
 * @param   ref               input reference set of frames
 * @param   test              input set of frames to test against reference
 *
 * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
cpl_error_code esotk_checkmatch(const cpl_frameset *ref, const cpl_frameset *test)
{
    const cpl_size N_ext       = cpl_frame_get_nextensions(cpl_frameset_get_position_const(ref, 0));
    const cpl_size N_ext_start = N_ext > 0 ? 1 : 0;

    const cpl_size N_frames = cpl_frameset_get_size(ref);
    for (cpl_size i = 0; i < N_frames; i++) {

        const cpl_frame *cref = cpl_frameset_get_position_const(ref , i);
        const cpl_frame *ctst = cpl_frameset_get_position_const(test, i);

        if (cpl_frame_get_nextensions(ctst) != cpl_frame_get_nextensions(cref)) {

            /* Check that we are not clobbering an existing error code. */
            cpl_error_code errorcode = cpl_error_get_code();
            if (errorcode == CPL_ERROR_NONE) {
                errorcode = CPL_ERROR_INCOMPATIBLE_INPUT;
            }

            return cpl_error_set_message(cpl_func, errorcode,
				 "'%s' mismatches no. of extensions of '%s'",
				 cpl_frame_get_filename(ctst), cpl_frame_get_filename(cref));
        }

        for (cpl_size ext = N_ext_start; ext <= N_ext; ext++) {

            cpl_propertylist *ref_plist = cpl_propertylist_load(cpl_frame_get_filename(cref), ext);
            cpl_propertylist *tst_plist = cpl_propertylist_load(cpl_frame_get_filename(ctst), ext);

            const cpl_size refx = cpl_propertylist_get_int(ref_plist, "NAXIS1");
            const cpl_size refy = cpl_propertylist_get_int(ref_plist, "NAXIS2");
            const cpl_size tstx = cpl_propertylist_get_int(tst_plist, "NAXIS1");
            const cpl_size tsty = cpl_propertylist_get_int(tst_plist, "NAXIS2");

            cpl_propertylist_delete(ref_plist);
            cpl_propertylist_delete(tst_plist);

            if (tstx != refx || tsty != refy) {
                return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
					 "Optional file '%s' mismatches pixel dimensions of '%s'",
					 cpl_frame_get_filename(ctst), cpl_frame_get_filename(cref));
            } else {
                cpl_msg_debug(cpl_func, "Dimension checks passed on: '%s' vs '%s'",
                     cpl_frame_get_filename(cref), cpl_frame_get_filename(ctst));
            }
        }
    }

    return CPL_ERROR_NONE;
}

/**@}*/
