/*
 * This file is part of the HDRLDEMO pipeline
 * Copyright (C) 2022 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#define _DEFAULT_SOURCE
#define _BSD_SOURCE

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------*/
/**
 *                              Defines
 */
/*----------------------------------------------------------------------------*/

#define RECIPE_NAME                                         "esotk_barycorr"

/*----------------------------------------------------------------------------*/
/**
 *                              Includes
 */
/*----------------------------------------------------------------------------*/

#include <cpl.h>
#include "hdrl.h"
#include "esotk_dfs.h"
#include "esotk_utils.h"


/*----------------------------------------------------------------------------*/
/**
 *                 Functions prototypes
 */
/*----------------------------------------------------------------------------*/


static cpl_error_code
esotk_barycorr_adjust_header(cpl_propertylist * header,
			     const char * keyword, double barycorr);

/*----------------------------------------------------------------------------*/
/**
 *                          Static variables
 */
/*----------------------------------------------------------------------------*/


static char esotk_barycorr_description[] =
"The recipe derives and on request (--barycorr.apply=TRUE and SPECSYS                        \n"
"== TOPOCENT or does not exist) tries to apply the barycentric                               \n"
"correction to data stored in table format. The extension containing                         \n"
"the table and the column name containing the wavelength can be set by                       \n"
"the recipe parameter (--barycorr.extension) and                                             \n"
"(--barycorr.colname-wave), accordingly.                                                     \n"
"                                                                                            \n"
"If the barycentric correction can be applied to the data the relevant                       \n"
"header parameter are adjusted for the applied barycentric correction.                       \n"
"                                                                                            \n"
"The derived barycentric correction is always stored in the (HIERARCH                        \n"
"ESO DRS BARYCORR) product header parameter. If the application of the                       \n"
"barycentric correction is requested, but it fails (e.g., the specified                      \n"
"column is missing or the input is not a table), the product will contain                    \n"
" only a primary header.                                                                     \n"
"                                                                                            \n"
"The algorithm do derive the barycentric correction is based on the                          \n"
"erfa library (https://github.com/liberfa/erfa). The relevant input is read                  \n"
"from the header parameter of the input files (RAW) by specifying the keyword                \n"
"as follows:                                                                                 \n"
"                                                                                            \n"
"--RECIPE_PARAMETER:                                                                         \n"
"  RECIPE_PARAMETER_DESCRIPTION [default recipe value]                                       \n"
"  ADDITIONAL_INFO              [unit]                                                       \n"
"                                                                                            \n"
" --barycorr.hname-ra:                                                                       \n"
"   Name of the header keyword containing the right ascension. [RA]                          \n"
"   Target right ascension (J2000) [deg]                                                     \n"
"                                                                                            \n"
"--barycorr.hname-dec:                                                                       \n"
"   Name of the header keyword containing the declination. [DEC]                             \n"
"   Target declination (J2000) [deg]                                                         \n"
"                                                                                            \n"
"--barycorr.hname-mjdobs:                                                                    \n"
"  Name of the header keyword containing the mjd-obs. [MJD-OBS]                              \n"
"  Start of observation [days]                                                               \n"
"                                                                                            \n"
"--barycorr.hname-exptime:                                                                   \n"
"  Name of the header keyword containing the exposure time. [EXPTIME]                        \n"
"  Time to mid exposure - half of the value added to MJD-OBS. [s]                            \n"
"                                                                                            \n"
"--barycorr.hname-longitude:                                                                 \n"
"  Name of the header keyword containing the longitude of the observatory. [ESO TEL GEOLON]  \n"
"  Telescope geodetic longitude (+ = East ) [deg]                                            \n"
"                                                                                            \n"
"--barycorr.hname-latitude:                                                                  \n"
"  Name of the header keyword containing the latitude of the observatory. [ESO TEL GEOLAT]   \n"
"  Telescope geodetic latitude  (+ = North) [deg]                                            \n"
"                                                                                            \n"
"--barycorr.hname-elevation:                                                                 \n"
"  Name of the header keyword containing the elevation of the observatory. [ESO TEL GEOELEV] \n"
"  Telescope elevation above sea level [m]                                                   \n"
"                                                                                            \n"
"--barycorr.pressure:                                                                        \n"
"  Pressure during observation. [0.0]                                                        \n"
"  Pressure at the observer [hPa == mbar]                                                    \n"
"                                                                                            \n"
"--barycorr.temperature:                                                                     \n"
"  Temperature during observation. [0.0]                                                     \n"
"  Ambient temperature at the observer [deg C]                                               \n"
"                                                                                            \n"
"--barycorr.humidity:                                                                        \n"
"  Humidity during observation. [0.0]                                                        \n"
"  Relative humidity at the observer [range 0 - 1]                                           \n"
"                                                                                            \n"
"--barycorr.wavelength:                                                                      \n"
"  Wavelength during observation. [0.0]                                                      \n"
"  Observing wavelength [micrometer]                                                         \n"
"                                                                                            \n"
"Please note that the pressure, temperature, humidity, and the                               \n"
"wavelength parameters are currently always kept to a value of 0. No tests with              \n"
"other values were performed.                                                                \n"
"                                                                                            \n"
"Input files:                                                                                \n"
"                                                                                            \n"
"   DO category:                 Explanation:                       Required:                \n"
"   --------------------------   --------------------------------   ---------                \n"
"   RAW                          Input file(s) (1 to N)             Yes                      \n"
"   EOP_PARAM                    Earth Orientation Parameter file   Yes                      \n"
"                                                                                            \n"
"Output files:                                                                               \n"
"                                                                                            \n"
"   DO category:                 Explanation:                                                \n"
"   --------------------------   --------                                                    \n"
"   ESOTK_BARYCORR               Barycentric corrected file and/or header keyword            \n";

/* Standard CPL recipe definition */
cpl_recipe_define( esotk_barycorr,
		   ESOTK_BINARY_VERSION,
		   "HDRL Group",
		   PACKAGE_BUGREPORT,
		   "2022",
		   "Derives and applies the barrycentric correction",
		   esotk_barycorr_description);

/*----------------------------------------------------------------------------*/
/**
 *                 Functions
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_barycorr Derives the barycentric correction of an observation
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @brief   derives the barycentric correction of an observation
 *
 * @param   frameset   input set of frames
 * @param   parlist    input recipe parameters
 *
 * @return  cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static int esotk_barycorr(
    cpl_frameset            *frameset,
    const cpl_parameterlist *parlist)
{
  if (esotk_check_and_set_groups(frameset) != CPL_ERROR_NONE) {
      return cpl_error_get_code();
  }

  /* Temporary framesets */
  cpl_size     fsize         = cpl_frameset_get_size(frameset);
  cpl_frameset *frameset_raw = cpl_frameset_new();
  cpl_frame    *eop_frame    = NULL;


  /* Split the input frameset  */
  for (cpl_size i = 0; i < fsize; i++) {

      cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);

      if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW)) {
	  cpl_frameset_insert(frameset_raw, cpl_frame_duplicate(cur_frame));
	  cpl_msg_info(cpl_func,"%s frame found in the SOF: %s", ESOTK_RAW,
		       cpl_frame_get_filename(cur_frame));
      } else if (eop_frame == NULL &&
	  !strcmp(ESOTK_EOP_PARAM, cpl_frame_get_tag(cur_frame))) {
	  eop_frame = cpl_frame_duplicate(cur_frame);
	  cpl_msg_info(cpl_func,"%s frame found in the SOF: %s", ESOTK_EOP_PARAM,
		       cpl_frame_get_filename(cur_frame));
      } else {
	  cpl_msg_info(cpl_func, "SOF contains unknown tag: %s",
		       cpl_frame_get_tag(cur_frame));
      }
  }

  /* Check if exist one or more raw frames */
  if (cpl_frameset_get_size(frameset_raw) < 1) {
      cpl_frameset_delete(frameset_raw);
      cpl_frame_delete(eop_frame);
      return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
				   "Frames of type RAW missing");
  }

  /* Check if exist one EOP frame */
  if (eop_frame == NULL) {
      cpl_frameset_delete(frameset_raw);
      return cpl_error_set_message(cpl_func, CPL_ERROR_INCOMPATIBLE_INPUT,
				   "Frame of type EOP_PARAM missing ");
  }

  double  ra = 0., dec = 0., mjdobs = 0., exptime = 0., longitude = 0.;
  double  latitude = 0., elevation = 0., pressure = 0., temperature = 0.;
  double  humidity = 0., wavelength = 0.;


  /*** Get input parameter from the cpl_paramterlist ***/
  const cpl_parameter *par = NULL;

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-ra");
  const char*  hname_ra = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-dec");
  const char*  hname_dec = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-mjdobs");
  const char*  hname_mjdobs = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-exptime");
  const char*  hname_exptime = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-longitude");
  const char*  hname_longitude = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-latitude");
  const char*  hname_latitude = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.hname-elevation");
  const char*  hname_elevation = cpl_parameter_get_string(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.pressure");
  pressure = cpl_parameter_get_double(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.temperature");
  temperature = cpl_parameter_get_double(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.humidity");
  humidity = cpl_parameter_get_double(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.wavelength");
  wavelength = cpl_parameter_get_double(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.apply");
  const cpl_boolean  barycorr_apply  = cpl_parameter_get_bool(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.extension");
  const int   extension = cpl_parameter_get_int(par);

  par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".barycorr.colname-wave");
  const char*         colname_wave = cpl_parameter_get_string(par);


  /* Load EOP file */
  cpl_msg_info(cpl_func, "Reading table from %s frame %s", ESOTK_EOP_PARAM,
	       cpl_frame_get_filename(eop_frame) );
  cpl_table * eop_table = cpl_table_load(cpl_frame_get_filename(eop_frame), 1, 0);
  if (eop_table == NULL) {
      cpl_frameset_delete(frameset_raw);
      cpl_frame_delete(eop_frame);
      return cpl_error_get_code();
  }

  /* Doing some baisc checks on the eop table */
  if (!cpl_table_has_column(eop_table, "MJD") ||
      !cpl_table_has_column(eop_table, "PMX") ||
      !cpl_table_has_column(eop_table, "PMY") ||
      !cpl_table_has_column(eop_table, "DUT")){
      return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
				   "The %s table does not have all required "
				   "columns, i.e. MJD, PMX, PMY, DUT",
				   ESOTK_EOP_PARAM);
  }

  cpl_size fsize_raw  = cpl_frameset_get_size(frameset_raw);

  for (cpl_size i = 0; i < fsize_raw; i++) {
      cpl_errorstate  prestate = cpl_errorstate_get();

      cpl_frame *cur_frame = cpl_frameset_get_position(frameset_raw, i);
      cpl_msg_info(cpl_func, "Working on frame %s ", cpl_frame_get_filename(cur_frame) );
      cpl_msg_indent_more();
       /* Read the primary header */
      cpl_msg_info(cpl_func, "Reading the primary header ...");
      cpl_propertylist * header = NULL;
      header = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
      if (header == NULL) {
	  cpl_errorstate_set(prestate);
	  cpl_msg_warning(cpl_func, "Can not load primary header. Frame %s will "
	      "be skipped", cpl_frame_get_filename(cur_frame));
	  cpl_msg_indent_less();
	  continue;
      }

      /* Removing RADECSYS from primary header as renamed to RADESYS by cpl */
      cpl_propertylist_erase_regexp(header, "RADECSYS", 0);

      /* Trying to load a table in case we have an IDP
       * If the loading fails cpl_table_load will return NULL and we reset */
      cpl_msg_info(cpl_func, "Reading the table containing the data ...");
      prestate = cpl_errorstate_get();
      cpl_table * table = NULL;
      cpl_propertylist * table_header = NULL;
      table = cpl_table_load(cpl_frame_get_filename(cur_frame), extension, 0);
      table_header = cpl_propertylist_load(cpl_frame_get_filename(cur_frame),
					   extension);
      if (table == NULL) {
	  cpl_errorstate_set(prestate);
	  cpl_msg_warning(cpl_func, "Can not load the table containing the data."
	      " The recipe product of frame %s will only contain a primary header"
	      " and no data section", cpl_frame_get_filename(cur_frame));
      }

      cpl_errorstate_set(prestate);

      ra        = cpl_propertylist_get_double(header, hname_ra       );
      dec       = cpl_propertylist_get_double(header, hname_dec      );
      mjdobs    = cpl_propertylist_get_double(header, hname_mjdobs   );
      exptime   = cpl_propertylist_get_double(header, hname_exptime  );
      longitude = cpl_propertylist_get_double(header, hname_longitude);
      latitude  = cpl_propertylist_get_double(header, hname_latitude );
      elevation = cpl_propertylist_get_double(header, hname_elevation);

      if (!cpl_errorstate_is_equal(prestate)) {
	  cpl_msg_warning(cpl_func, "Can not read one or more requested keywords"
	      " (as specified by the recipe parameters) from the primary header."
	      " Frame %s will be skipped", cpl_frame_get_filename(cur_frame));
	  if (header != NULL) cpl_propertylist_delete(header);
	  if (table_header != NULL) cpl_propertylist_delete(table_header);
	  if (table != NULL)  cpl_table_delete(table);
	  cpl_errorstate_set(prestate);
	  cpl_msg_indent_less();
	  continue;
      }

      double barycorr = 0;
      cpl_error_code err = CPL_ERROR_NONE;
      err = hdrl_barycorr_compute(ra, dec, eop_table, mjdobs, exptime/2.,
				   longitude,  latitude, elevation,
				   pressure, temperature, humidity, wavelength,
				   &barycorr);

      if(err != CPL_ERROR_NONE) {
	  cpl_msg_warning(cpl_func, "ERFA can not compute the barycentric "
			  "correction for this frame. Frame %s will be skipped",
			  cpl_frame_get_filename(cur_frame));
	  if (header != NULL) cpl_propertylist_delete(header);
	  if (table_header != NULL) cpl_propertylist_delete(table_header);
	  if (table != NULL)  cpl_table_delete(table);
	  cpl_errorstate_set(prestate);
	  cpl_msg_indent_less();
	  continue;

      }

      cpl_msg_info (cpl_func, "Barycentric correction: %g km/s",
		    barycorr/1000.);
      cpl_msg_info (cpl_func, "Writing the barycentric correction to the primary"
		    " header as ESO.DRS.BARYCORR");

      char * outfile = NULL ;
      outfile = cpl_sprintf("esotk_barycorr_%04d.fits", (int)i);
      /* Add a QC parameter  */

      /* Add the product category and save image */
      cpl_propertylist_update_string(header, CPL_DFS_PRO_CATG, "ESOTK_BARYCORR");

      cpl_propertylist_update_double(header,"ESO DRS BARYCORR", barycorr/1000);
      cpl_propertylist_set_comment(header,"ESO DRS BARYCORR", "Barycentric "
	  "correction [km/s]");

      /* Check header for SPECSYS */
      char * specsys = NULL;
      if (cpl_propertylist_has(header, "SPECSYS")) {
	  specsys = cpl_sprintf("%s",
				cpl_propertylist_get_string(header, "SPECSYS"));
      } else {
	  cpl_msg_warning(cpl_func, "No header keyword SPECSYS found. "
	      "Assuming TOPOCENT");
	  specsys = cpl_sprintf("%s",  "TOPOCENT");
      }

      if (table != NULL && cpl_table_has_column (table, colname_wave) &&
	  !strcmp(specsys, "TOPOCENT") && barycorr_apply == CPL_TRUE) {
	  cpl_msg_info(cpl_func, "Updating SPECSYS keyword in the primary header");
	  cpl_propertylist_update_string(header, "SPECSYS", "BARYCENT");

	  /* Adjust IDP header keywords */
	  cpl_msg_info(cpl_func, "Adjusting primary IDP header accordingly");
	  esotk_barycorr_adjust_header(header, "WAVELMIN", barycorr);
	  esotk_barycorr_adjust_header(header, "WAVELMAX", barycorr);
      }

      /* In order to inherit all parameters from the used frame only */
      cpl_frameset * frameset_singleframe = cpl_frameset_new();
      cpl_frameset_insert(frameset_singleframe, cpl_frame_duplicate(cur_frame));
      cpl_dfs_save_propertylist(frameset, NULL, parlist,
				frameset_singleframe, cur_frame,
				"esotk_barycorr", header, NULL,
				PACKAGE "/" PACKAGE_VERSION, outfile);
      cpl_frameset_delete(frameset_singleframe);

      /* Trying to save the table */
      if (table != NULL ) {
	  if (!strcmp(specsys, "TOPOCENT") &&
	      barycorr_apply == CPL_TRUE &&
	      cpl_table_has_column (table, colname_wave)) {
	      cpl_msg_info(cpl_func, "Applying barycentric correction to table");
	      cpl_size depth = cpl_table_get_column_depth(table, colname_wave);
	      if(depth > 0) { /* We are in the IDP format case */
		  cpl_array *wave_array = NULL;
		  wave_array = (cpl_array *)cpl_table_get_array (table, colname_wave, 0);
		  cpl_array_multiply_scalar(wave_array,
					    1 + barycorr / CPL_PHYS_C);
	      } else {
		  cpl_table_multiply_scalar(table, colname_wave,
					    1 + barycorr / CPL_PHYS_C);
	      }

	      /* Adjust IDP header keywords */
	      cpl_msg_info(cpl_func, "Adjusting some IDP table header keywords");
	      esotk_barycorr_adjust_header(table_header, "TDMIN1", barycorr);
	      esotk_barycorr_adjust_header(table_header, "TDMAX1", barycorr);
	      esotk_barycorr_adjust_header(table_header, "SPEC_VAL", barycorr);
	      esotk_barycorr_adjust_header(table_header, "SPEC_BW", barycorr);

	  }
	  cpl_msg_info(cpl_func, "Saving barycentric corrected table to final recipe product ...");
	  cpl_table_save(table, NULL, table_header, outfile, CPL_IO_EXTEND);
	  cpl_table_delete(table);
      }

      if (table_header != NULL) {
	  cpl_propertylist_delete(table_header);
      }
      cpl_free(specsys);
      cpl_free(outfile);

      cpl_propertylist_delete(header);
      cpl_msg_indent_less();
  }


  /*** Cleanup ***/
  cpl_frameset_delete(frameset_raw);
  cpl_frame_delete(eop_frame);
  cpl_table_delete(eop_table);

  return (int)cpl_error_get_code();
}

static cpl_error_code
esotk_barycorr_adjust_header(cpl_propertylist * header,
			     const char * keyword, double barycorr){
  /* Adjust IDP header keywords */
  if (cpl_propertylist_has(header, keyword)) {
      double value = cpl_propertylist_get_double(header, keyword);
      value *= (1 + barycorr / CPL_PHYS_C);
      cpl_propertylist_update_double(header, keyword, value);
  }
  return cpl_error_get_code();
}


/** @cond PRIVATE */

/*----------------------------------------------------------------------------*/
/**
 * @brief   Function needed by cpl_recipe_define to fill the input parameters
 *
 * @param   self         Output parameterlist
 *
 * @return  cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_barycorr_fill_parameterlist(
    cpl_parameterlist *self)
{

  cpl_parameter *par;

  /* --user-specified header keywords  */

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-ra",	CPL_TYPE_STRING,
				" Name of the header keyword containing the right ascension",
				RECIPE_NAME, "RA");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-ra");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-dec",	CPL_TYPE_STRING,
				" Name of the header keyword containing the declination ",
				RECIPE_NAME, "DEC");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-dec");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-mjdobs",	CPL_TYPE_STRING,
				" Name of the header keyword containing the mjd-obs ",
				RECIPE_NAME, "MJD-OBS");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-mjdobs");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-exptime",	CPL_TYPE_STRING,
				" Name of the header keyword containing the exposure time ",
				RECIPE_NAME, "EXPTIME");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-exptime");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-longitude",	CPL_TYPE_STRING,
				" Name of the header keyword containing the longitude of the observatory",
				RECIPE_NAME, "ESO TEL GEOLON");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-longitude");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);


  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-latitude",	CPL_TYPE_STRING,
				" Name of the header keyword containing the latitude of the observatory",
				RECIPE_NAME, "ESO TEL GEOLAT");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-latitude");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.hname-elevation",	CPL_TYPE_STRING,
				" Name of the header keyword containing the elevation of the observatory",
				RECIPE_NAME, "ESO TEL GEOELEV");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.hname-elevation");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.pressure", CPL_TYPE_DOUBLE,
				"Pressure during observation",
				RECIPE_NAME, 0.);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.pressure");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.temperature", CPL_TYPE_DOUBLE,
				"Temperature during observation",
				RECIPE_NAME, 0.);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.temperature");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.humidity", CPL_TYPE_DOUBLE,
				"Humidity during observation",
				RECIPE_NAME, 0.);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.humidity");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.wavelength", CPL_TYPE_DOUBLE,
				"Wavelength during observation",
				RECIPE_NAME, 0.);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.wavelength");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.apply", CPL_TYPE_BOOL,
				"Apply barycentric correction to the wavelength column.",
				RECIPE_NAME, CPL_TRUE);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.apply");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* extension in case of no IDP convention */
  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.extension",	CPL_TYPE_INT,
				"Fits extension containing the spectrum table",
				RECIPE_NAME, 1);
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.extension");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);

  /* name of the column containing the wavelength in case of no IDP convention */
  par = cpl_parameter_new_value(RECIPE_NAME".barycorr.colname-wave",	CPL_TYPE_STRING,
				" Name of the table column containing the wavelength "
				"to be corrected ",
				RECIPE_NAME, "WAVE");
  cpl_parameter_set_alias( par, CPL_PARAMETER_MODE_CLI, "barycorr.colname-wave");
  cpl_parameter_disable(   par, CPL_PARAMETER_MODE_ENV);
  cpl_parameterlist_append(self, par);



  return CPL_ERROR_NONE;
}


/** @endcond */


/**@}*/
