/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                    Includes
 -----------------------------------------------------------------------------*/

#include "esotk_utils.h"
#include "esotk_dfs.h"

#include "hdrl.h"

#include <cpl.h>

/*----------------------------------------------------------------------------*/
/**
  @defgroup object_mask computation recipe
 */
/*----------------------------------------------------------------------------*/

/**@{*/


/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

/* save the created mask */
static cpl_error_code esotk_mask_create_save(
        const cpl_mask          *objmask_filtered,
        const cpl_parameterlist *parlist,
        cpl_frameset            *frameset,
        const cpl_propertylist  *qclist,
        cpl_boolean             is_newframe,
        cpl_size                N_ext,
        cpl_size                frame);

/*-----------------------------------------------------------------------------
                            Static global variables
 -----------------------------------------------------------------------------*/

#define RECIPE_NAME "esotk_mask_create"

static char esotk_mask_create_description[] =
"                                                                           \n"
"The recipe derives an object mask from an astronomical FITS image,         \n"
"i.e. all pixels belonging to detected sources are flagged. Objects are     \n"
"detected and parameterised using the processed images and confidence       \n"
"maps.                                                                      \n"
"                                                                           \n"
"Input files:                                                               \n"
"                                                                           \n"
"  DO category:                 Explanation:           Required:            \n"
"  RAW                          Input Data             Yes                  \n"
"  RAW_BPM                      Bad Pixel Map          No                   \n"
"  RAW_CONFMAP                  Confidence Map         No                   \n"
"                                                                           \n"
"Output files:                                                              \n"
"                                                                           \n"
"  DO category:                 Explanation:                                \n"
"  OBJ_MASK                     Object mask                                 \n"
"                                                                           \n"
"                                                                           \n"
"Usage of the recipe:                                                       \n"
"                                                                           \n"
"In brief, the recipe first estimates the local sky background over the     \n"
"field and track any variations at adequate resolution to eventually        \n"
"remove them (if bkg.estimate=TRUE). Then it detects objects/blends of      \n"
"objects and keep a list of pixels belonging to each blend. Finally, it     \n"
"flaggs all pixels belonging to the detected objects.                       \n"
"                                                                           \n"
"Note that pixels are weighted in the source detection according to         \n"
"their value in the confidence map.                                         \n"
"                                                                           \n"
"                                                                           \n"
"Recipe parameter to characterize the detector:                             \n"
"                                                                           \n"
"--det.effective-gain: Effective detector gain value to convert             \n"
"                      intensity to electrons.                              \n"
"--det.saturation:     Detector saturation value. All pixel values          \n"
"                      exceeding this threshold before background           \n"
"                      subtraction are marked as bad pixels.                \n"
"                                                                           \n"
"Recipe parameter controlling the background determination and              \n"
"subtraction:                                                               \n"
"                                                                           \n"
"--bkg.estimate:  If set, the recipe estimates and subtracts the            \n"
"                 background from the input image, else, no background      \n"
"                 determination and subtraction is done.                    \n"
"                                                                           \n"
"--bkg.mesh-size: It defines the mesh size in pixels to construct the       \n"
"                 low resolution background by removing the objects. If     \n"
"                 the mesh-size is to small, object residuals may           \n"
"                 distort the background.                                   \n"
"                                                                           \n"
"--bkg.smooth-gauss-fwhm: The FWHM of the Gaussian kernel used to           \n"
"                         convolve the background before doing object       \n"
"                         detection.                                        \n"
"                                                                           \n"
"Recipe parameter controlling the object detection:                         \n"
"                                                                           \n"
"                                                                           \n"
"--obj.min-pixels:  Minimum pixel area (contiguous) of an object to be      \n"
"                   detected.                                               \n"
"--obj.threshold:   Minimum detection threshold in sigma above sky for a    \n"
"                   pixel to be considered for detection.                   \n"
"--obj.deblending:  If set, the algorithm tries to deblend close-by         \n"
"                   objects.                                                \n"
"--obj.core-radius: Value of the core radius. The core radius, among        \n"
"                   others, determines the fixed aperture flux              \n"
"                   measurements - it uses the core radius multiplied       \n"
"                   by predefined factor.                                   \n"
"                                                                           \n"
"The derived bad pixel mask is also filtered by a CLOSING or OPENING        \n"
"filter. The filtering process can be controlled by the parameters          \n"
"(--pfx), (--pfy), and (--pfm):                                             \n"
"                                                                           \n"
"--pfx: X Size in pixel of the post filtering kernel.                       \n"
"--pfy: Y Size in pixel of the post filtering kernel.                       \n"
"--pfm: Post filtering mode. <closing | dilation>                           \n"
"                                                                           \n"
"If pfx and pfy are set to 0, no post-processing is applied.                \n"
"                                                                           \n"
"Please note that part of the code is paralelised. In order to optimise     \n"
"use of computing resources you should set the environment variable         \n"
"OMP_NUM_THREADS to a proper value, like (for bash), to use 4 cores         \n"
"export OMP_NUM_THREADS=4                                                   \n";


/* Standard CPL recipe definition */
cpl_recipe_define( esotk_mask_create,
                   ESOTK_BINARY_VERSION,
                   "HDRL Group",
                   PACKAGE_BUGREPORT,
                   "2016",
                   "Creates an object mask",
                   esotk_mask_create_description);

/*----------------------------------------------------------------------------*/
/**
 * @brief Function needed by cpl_recipe_define to fill the input parameters
 *
 * @param  self   parameterlist where you need put parameters
 *
 * @return (int)cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_mask_create_fill_parameterlist(cpl_parameterlist *self)
{
	/* Default values of the hdrl_parameter in the hdrl_catalogue */
	int                    obj_min_pixels  = 4;
    double                 obj_threshold   = 2.5;
    cpl_boolean            obj_deblending  = CPL_TRUE;
    double                 obj_core_radius = 5.;
    cpl_boolean            bkg_estimate    = CPL_TRUE;
    int                    bkg_mesh_size   = 64;
    double                 bkg_smooth_fwhm = 2.;
    double                 det_eff_gain    = 3.;
    double                 det_saturation  = HDRL_SATURATION_INIT;
    hdrl_catalogue_options resulttype      = HDRL_CATALOGUE_BKG;


    /* Create a parameter list */
    hdrl_parameter *c_def = hdrl_catalogue_parameter_create(
    	obj_min_pixels, obj_threshold, obj_deblending, obj_core_radius,
		bkg_estimate, bkg_mesh_size, bkg_smooth_fwhm,
		det_eff_gain, det_saturation, resulttype);


    cpl_parameterlist *s_param = hdrl_catalogue_parameter_create_parlist(RECIPE_NAME, "", c_def);
    hdrl_parameter_delete(c_def) ;
    
    /* Duplicate in the input/output parameterlist the parameters in the catalogue */
    for (cpl_parameter *p = cpl_parameterlist_get_first(s_param); p != NULL;
    	p = cpl_parameterlist_get_next(s_param) ) {
    	cpl_parameterlist_append(self, cpl_parameter_duplicate(p));
    }
    cpl_parameterlist_delete(s_param);


    /* Get default parameters in the recipe */
    cpl_parameter *par;

    /* --hdrldemo_bpm_2d.post-filter-x */
    par = cpl_parameter_new_value(RECIPE_NAME".post-filter-x", CPL_TYPE_INT,
            "X Size of the post filtering kernel.", RECIPE_NAME, 3);
    cpl_parameter_set_alias(par, CPL_PARAMETER_MODE_CLI, "pfx");
    cpl_parameter_disable(par, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(self, par);

    /* --hdrldemo_bpm_2d.post-filter-y */
    par = cpl_parameter_new_value(RECIPE_NAME".post-filter-y", CPL_TYPE_INT,
            "Y Size of the post filtering kernel.", RECIPE_NAME, 3);
    cpl_parameter_set_alias(par, CPL_PARAMETER_MODE_CLI, "pfy");
    cpl_parameter_disable(par, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(self, par);

    /* --hdrldemo_bpm_2d.post-filter-mode */
    par = cpl_parameter_new_enum(RECIPE_NAME".post-filter-mode", 
            CPL_TYPE_STRING, "Post filtering mode.", RECIPE_NAME,
            "closing", 2, "closing", "dilation");
    cpl_parameter_set_alias(par, CPL_PARAMETER_MODE_CLI, "pfm");
    cpl_parameter_disable(par, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(self, par);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief Recipe for create a mask
 *
 * @param  frameset   input set of frames
 * @param  parlist    input recipe parameters
 *
 * @return (int)cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static int esotk_mask_create(
    cpl_frameset            *frameset,
    const cpl_parameterlist *parlist)
{
    if (esotk_check_and_set_groups(frameset) != CPL_ERROR_NONE) {
    	return cpl_error_get_code();
    }

    /* Get parameters */
    const cpl_parameter *par;

    par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".post-filter-x");
    const int pfx = cpl_parameter_get_int(par);

    par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".post-filter-y");
    const int pfy = cpl_parameter_get_int(par);

    par = cpl_parameterlist_find_const(parlist, RECIPE_NAME".post-filter-mode");
    const char *pfm = cpl_parameter_get_string(par); /* Only pointer, not store memory */

    cpl_filter_mode filter_mode;
    if(!strcmp(pfm, "closing")) {
        filter_mode = CPL_FILTER_CLOSING ;
    } else if(!strcmp(pfm, "dilation")) {
        filter_mode = CPL_FILTER_DILATION ;
    } else {
        return cpl_error_set_message(cpl_func, CPL_ERROR_ILLEGAL_INPUT,
                "Filter mode can only be \"closing\" or \"dilation\" (not %s)", pfm);
    }
    
    /* catalogue currently is always created */
    hdrl_parameter *p = hdrl_catalogue_parameter_parse_parlist(parlist, RECIPE_NAME);
    const hdrl_catalogue_options opt = HDRL_CATALOGUE_SEGMAP;
    hdrl_catalogue_parameter_set_option(p, opt);

    cpl_msg_info(cpl_func, "Classifying input frames:");

    /* Create the frameset for the different frame types */
    cpl_frameset *img_frameset = cpl_frameset_new();
    cpl_frameset *cnf_frameset = cpl_frameset_new();
    cpl_frameset *bpm_frameset = cpl_frameset_new();

    cpl_size N_fs = cpl_frameset_get_size(frameset);
    for (cpl_size i = 0; i < N_fs; i++){

    	/* Only pointers, not store memory */
        const cpl_frame *cur_frame = cpl_frameset_get_position_const(frameset, i);
        const char      *cur_tag   = cpl_frame_get_tag(cur_frame);

        if (!strcmp(cur_tag, ESOTK_RAW)){
            cpl_frameset_insert(img_frameset, cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cur_tag, ESOTK_RAW_CONFMAP)) {
            cpl_frameset_insert(cnf_frameset, cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cur_tag, ESOTK_RAW_BPM)) {
            cpl_frameset_insert(bpm_frameset, cpl_frame_duplicate(cur_frame));
        } else {
            cpl_msg_warning(cpl_func, "%s: Unsupported tag %s - ignoring data.",
                cpl_frame_get_filename(cur_frame), cpl_frame_get_tag(cur_frame));
        }
    }

    /* Check integrity */
    if (esotk_integrity_checks(img_frameset, cnf_frameset, bpm_frameset,
		 NULL, NULL, NULL, NULL, NULL, NULL, NULL, p) != CPL_ERROR_NONE) {
		return (int)cpl_error_get_code();
	}


    /* Count frames and report */
    cpl_size N_img = cpl_frameset_get_size(img_frameset);
    cpl_size N_cnf = cpl_frameset_get_size(cnf_frameset);
    cpl_size N_bpm = cpl_frameset_get_size(bpm_frameset);
    cpl_size N_ext = cpl_frame_get_nextensions(cpl_frameset_get_position(frameset, 0));


    /* If there are multi extension frames the primary extension is empty */
    cpl_size N_ext_start = N_ext > 0 ? 1 : 0;

    /* Detection and mask creation */
    cpl_msg_info(cpl_func, "Starting object detection and mask creation:");

    for (cpl_size frame = 0; frame < N_img; frame++) {

        cpl_boolean is_newframe = CPL_TRUE;

        /* Only a pointer, not new storage memory */
        cpl_frame *img_frame = cpl_frameset_get_position(img_frameset, frame);

        /* load images */
        for (cpl_size extension = N_ext_start; extension <= N_ext; extension++) {

            cpl_msg_debug(cpl_func, "frame = %d, extension = %d", (int)frame, (int)extension);

            cpl_image *img = cpl_image_load(cpl_frame_get_filename(img_frame), CPL_TYPE_FLOAT, 0, extension);
            if (cpl_error_get_code() == CPL_ERROR_NONE && N_bpm > 0) {

                /* Only a pointer, not new storage memory */
            	cpl_frame *bpm_frame = cpl_frameset_get_position(bpm_frameset, frame);

                cpl_image *bpm_image = cpl_image_load(cpl_frame_get_filename(bpm_frame), CPL_TYPE_FLOAT, 0, extension);
                if (cpl_error_get_code() == CPL_ERROR_NONE) {

                    cpl_mask *bpm = cpl_mask_threshold_image_create(bpm_image, -0.5, 0.5);
                    cpl_mask_not(bpm);

                    cpl_msg_debug(cpl_func, "extension: %d, frame: %d, Number of Bad pixels: %d",
                    		(int)extension, (int)frame, (int)cpl_mask_count(bpm));

                    cpl_image_reject_from_mask(img, bpm);

                    cpl_mask_delete(bpm);
                }
                if (bpm_image) cpl_image_delete(bpm_image);
            }

            cpl_image *cnf = NULL;
            if (N_cnf > 0 && cpl_error_get_code() == CPL_ERROR_NONE) {

            	/* Only a pointer, not new storage memory */
            	cpl_frame *cnf_frame = cpl_frameset_get_position(cnf_frameset, frame);

                cnf = cpl_image_load(cpl_frame_get_filename(cnf_frame), CPL_TYPE_FLOAT, 0, extension);

            } else if (cpl_error_get_code() == CPL_ERROR_NONE) {
                cpl_msg_info(cpl_func, "%s", "Missing confidence file tagged as "
                                ESOTK_RAW_CONFMAP", assuming unity");
            }

            if (cpl_error_get_code() == CPL_ERROR_NONE) {

                cpl_msg_info(cpl_func, "Working on extension %"CPL_SIZE_FORMAT""
                                " of frame %"CPL_SIZE_FORMAT, extension, frame);

                /* Calling the hdrl catalogue module */
                hdrl_catalogue_result *res = hdrl_catalogue_compute(img, cnf, NULL, p);

                if (cpl_error_get_code() == CPL_ERROR_NONE) {

                    cpl_msg_indent_more();
                    long detected_objects = (long)cpl_image_get_max(res->segmentation_map);
                    cpl_msg_info(cpl_func, "%ld detected objects", detected_objects);
                    cpl_msg_indent_less();
    
                    cpl_mask *objmask = cpl_mask_threshold_image_create(res->segmentation_map, -0.5, 0.5);
                    cpl_mask_not(objmask);

                    /* Post Filtering */
                    cpl_mask *objmask_filtered; 
                    if (pfx > 0 && pfy > 0) {
                        objmask_filtered = hdrl_bpm_filter(objmask, pfx, pfy, filter_mode);
                        cpl_mask_delete(objmask);
                    } else {
                        objmask_filtered = objmask;
                    }

                    cpl_propertylist *qclist = cpl_propertylist_new();
                    cpl_propertylist_update_long(qclist, "ESO QC NDETECTED", detected_objects);
                    cpl_propertylist_set_comment(qclist, "ESO QC NDETECTED", "total number of detected sources");

                    esotk_mask_create_save(objmask_filtered, parlist, frameset, qclist, is_newframe, N_ext, frame);

                    cpl_propertylist_delete(qclist);
                    cpl_mask_delete(objmask_filtered);
                    hdrl_catalogue_result_delete(res);

                } else {

                    if (img) cpl_image_delete(img);
                    if (cnf) cpl_image_delete(cnf);
                    hdrl_catalogue_result_delete(res);

                    esotk_clean_recipe_parameters(img_frameset, cnf_frameset, bpm_frameset,
                    		NULL, NULL, NULL, NULL, NULL, NULL, NULL, p);

                    return cpl_error_set_message(cpl_func, CPL_ERROR_UNSPECIFIED,
                        "Unable to compute catalogue.");
                }
            }

            is_newframe = CPL_FALSE;

            if (img) cpl_image_delete(img);
            if (cnf) cpl_image_delete(cnf);
        }

    }

    /* cleanup */
    esotk_clean_recipe_parameters(img_frameset, cnf_frameset, bpm_frameset,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, p);

    return (int)cpl_error_get_code();
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
  * @brief   save the created mask
  *
  * @param   objmask_filtered  input filtered object mask
  * @param   parlist           input recipe parameters
  * @param   frameset          input set of frames
  * @param   qclist            input table with QC parameters
  * @param   is_newframe       input flag signaling a new frame
  * @param   N_ext             input number of the current extension
  * @param   frame             input number of the current frame
  *
  * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_mask_create_save(
        const cpl_mask          *objmask_filtered,
        const cpl_parameterlist *parlist,
        cpl_frameset            *frameset,
        const cpl_propertylist  *qclist,
        cpl_boolean             is_newframe,
        cpl_size                N_ext,
        cpl_size                frame)
{
	/* Check mask isn't null */
    cpl_ensure_code(objmask_filtered, CPL_ERROR_NULL_INPUT);

    /* Create property list for save */
    cpl_propertylist *applist = cpl_propertylist_new();
    cpl_propertylist_update_string(applist, CPL_DFS_PRO_CATG, ESOTK_OBJ_MASK);

    /* Save property list */
    char *filestr = cpl_sprintf("esotk_object_mask_%04d.fits", (int)frame);
    if(is_newframe && N_ext > 0) {
        cpl_dfs_save_propertylist(frameset, applist, parlist, frameset, 
                                  NULL, "esotk_mask_create", applist, 
                                  NULL, PACKAGE "/" PACKAGE_VERSION,
                                  filestr);
    }

    /* save mask */
    cpl_propertylist_append(applist, qclist);
    cpl_image *out = cpl_image_new_from_mask(objmask_filtered);
    if (N_ext > 0) {
        cpl_image_save(out, filestr, CPL_TYPE_FLOAT, applist, CPL_IO_EXTEND);
    } else {
        cpl_dfs_save_image(frameset, NULL, parlist, frameset, NULL, out,
                           CPL_TYPE_FLOAT, "esotk_mask_create", applist, 
                           NULL, PACKAGE "/" PACKAGE_VERSION, filestr);
    }

    /* Clean up */
    cpl_free(filestr);
    cpl_image_delete(out);
    cpl_propertylist_delete(applist);
    
    return cpl_error_get_code();
}
