/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                    Includes
 -----------------------------------------------------------------------------*/

#include "esotk_utils.h"
#include "esotk_dfs.h"
#include "esotk_masterfringe_utils.h"

#define HDRL_USE_EXPERIMENTAL
#include "hdrl.h"

#include <cpl.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_masterfringe_correct  master fringe subtraction
 * @par Synopsis: TBD
 * @par Input frames: TBD
 * @par Output frames: TBD
 * @code
 * @endcode
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

/* read master fringe */
hdrl_image * esotk_masterfringe_correct_read(
		const cpl_size extension,
        cpl_frame      *frame_masterfringe_data,
        cpl_frame      *frame_masterfringe_bpm,
        cpl_frame      *frame_masterfringe_error);

/* save hdrl_imagelist of fringes */
static cpl_error_code esotk_masterfringe_correct_save(
        const hdrl_imagelist    *hlist_fringe,
        const cpl_parameterlist *parlist,
        cpl_frameset            *frameset,
        const cpl_frameset      *frameset_inherit,
        const cpl_table         *qctable,
        const cpl_size          idx,
        const cpl_size          N_ext_start);

/*-----------------------------------------------------------------------------
                            Static global variables
 -----------------------------------------------------------------------------*/

#define RECIPE_NAME "esotk_masterfringe_correct"

static char esotk_masterfringe_correct_description[] =
"                                                                           \n"
"The recipe subtracts a properly scaled master-fringe image from the        \n"
"input images                                                               \n"
"                                                                           \n"
"Input files:                                                               \n"
"                                                                           \n"
"  DO category:                      Explanation:           Required:       \n"
"  RAW                               Data                   Yes             \n"
"  RAW_BPM                           Bad pixel masks        NO              \n"
"  RAW_ERROR                         Error image            NO              \n"
"  RAW_VARIANCE                      Variance image         NO              \n"
"  OBJ_MASK                          Object masks           NO              \n"
"  FRINGE_MASK                       Fringe mask            NO              \n"
"  MASTER_FRINGE                     Master-fringe          YES             \n"
"  MASTER_FRINGE_BPM                 BPM of master-fringe   NO              \n"
"  MASTER_FRINGE_ERROR               Error of master-fringe NO              \n"
"                                                                           \n"
"Output files:                                                              \n"
"                                                                           \n"
"  DO category:                      Explanation:                           \n"
"                                                                           \n"
"  FRINGE_CORRECTED                  Fringe corrected images                \n"
"  FRINGE_CORRECTED_BPM              Bad Pixel Mask of the fringe corrected \n"
"                                    images                                 \n"
"  FRINGE_CORRECTED_ERROR            Error of the fringe corrected images   \n"
"                                                                           \n"
"The algorithm consists of two main steps: i) The recipe estimates the      \n"
"background and the fringe scaling-factor (amplitude) for the               \n"
"master-fringe image (MASTER_FRINGE) of each image tagged as RAW. The       \n"
"required offset and scaling factor is determined using a least-square      \n"
"fit of the master-fringe image to the RAW image.  ii) The properly         \n"
"scaled (if rescale=TRUE) master-fringe is then subtracted from each        \n"
"individual RAW frame.                                                      \n"
"                                                                           \n"
"Different masks controlling the algorithm in different stages can be       \n"
"passed to the recipe:                                                      \n"
"                                                                           \n"
"A) RAW_BPM: One for each RAW image. Bad pixel mask of each single          \n"
"input image. Each step in the algorithm takes this mask into account.      \n"
"                                                                           \n"
"B) OBJ_MASK: One for each RAW image. Images flagging the objects in        \n"
"the input image. These masks are only taken into account in step i),       \n"
"i.e. when computing the scaling factors between MASTER_FRINGE and RAW,     \n"
"but ignored for step ii).                                                  \n"
"                                                                           \n"
"C) FRINGE_MASK: Single mask. This mask is only taken into account in       \n"
"step i), i.e. when computing the scaling factors between MASTER_FRINGE     \n"
"and RAW, but ignored for step ii). Please note that if the fringe mask     \n"
"has mask-data only in the primary extension or only in the first           \n"
"extension but the raw frames are multi-extension files, the mask-data      \n"
"is applied to all extension. This allows the user to e.g. exclude the      \n"
"borders of all extensions in step i) without the need of creating a        \n"
"multi-extension mask-file.                                                 \n"
"                                                                           \n"
"Please note that part of the code is paralelised. In order to optimise     \n"
"use of computing resources you should set the environment variable         \n"
"OMP_NUM_THREADS to a proper value, like (for bash), to use 4 cores         \n"
"export OMP_NUM_THREADS=4                                                   \n"
"                                                                           \n";

/* Standard CPL recipe definition */
cpl_recipe_define( esotk_masterfringe_correct,
                   ESOTK_BINARY_VERSION,
                   "HDRL Group",
                   PACKAGE_BUGREPORT,
                   "2016",
                   "Subtracts a Master-Fringe",
                   esotk_masterfringe_correct_description);

/*----------------------------------------------------------------------------*/
/**
 * @brief Function needed by cpl_recipe_define to fill the input parameters
 *
 * @param  self   parameterlist where you need put parameters
 *
 * @return (int)cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_masterfringe_correct_fill_parameterlist(cpl_parameterlist * self)
{
    /* Master-fringe related parameters */

	/* --hdrldemo_catalogue.ext-nb-raw */
    cpl_parameter *par = cpl_parameter_new_value(
    		RECIPE_NAME".rescale", CPL_TYPE_BOOL,
			"measure fringe amplitude and rescale fringe correction before subtracting?.",
			RECIPE_NAME, CPL_TRUE);
    cpl_parameter_set_alias(par, CPL_PARAMETER_MODE_CLI, "rescale");
    cpl_parameter_disable(  par, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(self, par);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   perform master fringe combination
  @param   frameset   input set of frames
  @param   parlist    input recipe parameters
  @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
static int esotk_masterfringe_correct(
    cpl_frameset            *frameset,
		const cpl_parameterlist *parlist)
{
    if (esotk_check_and_set_groups(frameset) != CPL_ERROR_NONE) {
    	return cpl_error_get_code();
    }

    /* Get parameters */
    const cpl_parameter *par    = cpl_parameterlist_find_const(parlist, RECIPE_NAME".rescale");
    const cpl_boolean   rescale = cpl_parameter_get_bool(par);

    cpl_msg_info(cpl_func, "Classifying input frames:");

    cpl_frameset *frameset_fringe          = cpl_frameset_new();
    cpl_frameset *frameset_fringe_bpm      = cpl_frameset_new();
    cpl_frameset *frameset_fringe_obj      = cpl_frameset_new();
    cpl_frameset *frameset_fringe_error    = cpl_frameset_new();
    cpl_frameset *frameset_fringe_mask     = cpl_frameset_new();

    cpl_boolean  is_data_new               = CPL_TRUE;
    cpl_boolean  is_bpm_new                = CPL_TRUE;
    cpl_boolean  is_error_new              = CPL_TRUE;

    cpl_frame    *frame_masterfringe_data  = NULL;
    cpl_frame    *frame_masterfringe_error = NULL;
    cpl_frame    *frame_masterfringe_bpm   = NULL;

    cpl_size N_fs = cpl_frameset_get_size(frameset);
    for (cpl_size i = 0; i < N_fs; i++) {

        cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);

        if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW)) {
            cpl_frameset_insert(frameset_fringe,       cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW_BPM)) {
            cpl_frameset_insert(frameset_fringe_bpm,   cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW_ERROR)) {
            cpl_frameset_insert(frameset_fringe_error, cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW_VARIANCE)) {
        	/* Insert variance maps into the error frameset and convert them
        	 * into a proper error map when reading - see below */
            cpl_frameset_insert(frameset_fringe_error, cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_OBJ_MASK)) {
            cpl_frameset_insert(frameset_fringe_obj,   cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_FRINGE_MASK)) {
        	/* 'frameset_fringe_mask' only contains the 'STATIC_MASK' tag */
            cpl_frameset_insert(frameset_fringe_mask , cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_MASTER_FRINGE)       && is_data_new ) {
            frame_masterfringe_data = cpl_frame_duplicate(cur_frame);
            is_data_new = CPL_FALSE;
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_MASTER_FRINGE_ERROR) && is_error_new) {
            frame_masterfringe_error = cpl_frame_duplicate(cur_frame);
            is_error_new = CPL_FALSE;
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_MASTER_FRINGE_BPM)   && is_bpm_new  ) {
            frame_masterfringe_bpm = cpl_frame_duplicate(cur_frame);
            is_bpm_new = CPL_FALSE;
        }
    }

    /* Check if data is present
     * TODO: check if the other need to be present: error or bpm */
    if (frame_masterfringe_data == NULL) {
        esotk_clean_recipe_parameters(
			frameset_fringe, NULL, frameset_fringe_bpm,
			frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
			frame_masterfringe_data, frame_masterfringe_error, frame_masterfringe_bpm,
			NULL, NULL);
        return cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                "Cannot find frames tagged");
    }

    /* Check integrity */
    if (esotk_integrity_checks(
		 frameset_fringe, NULL, frameset_fringe_bpm,
		 frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
		 frame_masterfringe_data, frame_masterfringe_error, frame_masterfringe_bpm,
		 NULL, NULL) != CPL_ERROR_NONE) {
		return (int)cpl_error_get_code();
	}

    /* load fringe mask */
    cpl_mask *fringe_mask = esotk_masterfringe_load_fringe_mask(frameset_fringe_mask);
    cpl_boolean single_fringe = fringe_mask != NULL ? CPL_TRUE : CPL_FALSE;


    cpl_msg_info(cpl_func, "Starting fringe correction:");

    /* If there are multi extension frames the primary extension is empty */
    const cpl_size N_obj  = cpl_frameset_get_size(frameset_fringe_obj);
    const cpl_size N_mask = cpl_frameset_get_size(frameset_fringe_mask);
    const cpl_size N_ext  = cpl_frame_get_nextensions(cpl_frameset_get_position(frameset, 0));
    const cpl_size N_ext_start = N_ext > 0 ? 1 : 0;
    for (cpl_size extension = N_ext_start; extension <= N_ext; extension++) {

		cpl_msg_info(cpl_func, "Working on extension %"CPL_SIZE_FORMAT, extension);

        hdrl_image *himg_masterfringe = esotk_masterfringe_correct_read(extension,
        	frame_masterfringe_data, frame_masterfringe_bpm, frame_masterfringe_error);

        if (himg_masterfringe) {

            /* no single global fringe, load fringe mask of extension */
			if (N_mask >= 1 && fringe_mask == NULL) {
				cpl_frame  *f_mask = cpl_frameset_get_position(frameset_fringe_mask, 0);
				const char *fn     = cpl_frame_get_filename(f_mask);
				fringe_mask        = cpl_mask_load(fn, 0, extension);
			}

			/* Loop over the frames */
	        hdrl_imagelist *hlist_fringe = hdrl_imagelist_new();
	        cpl_imagelist  *clist_obj    = cpl_imagelist_new();
	        if (esotk_masterfringe_loop_img_frames(extension, hlist_fringe, clist_obj,
	        		frameset_fringe, frameset_fringe_bpm, frameset_fringe_obj,
					frameset_fringe_error) != CPL_ERROR_NONE) {
				esotk_masterfringe_clean_all_recipe_parameters(
					frameset_fringe, NULL, frameset_fringe_bpm,
					frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
					frame_masterfringe_data, frame_masterfringe_error, frame_masterfringe_bpm,
					fringe_mask, NULL,
					himg_masterfringe, hlist_fringe, clist_obj, NULL, NULL, NULL);
	            return (int)cpl_error_get_code();
	        }

	        /* Correct hlist_fringe with the himg_masterfringe */
			cpl_table *qctable = NULL;
			if(rescale == CPL_FALSE) {
				hdrl_imagelist_sub_image(hlist_fringe, himg_masterfringe);
			} else if (N_obj > 0) {
				hdrl_fringe_correct(hlist_fringe, clist_obj, fringe_mask, himg_masterfringe, &qctable);
			} else {
				hdrl_fringe_correct(hlist_fringe, NULL,      fringe_mask, himg_masterfringe, &qctable);
			}

			/* Check correction */
			if (cpl_error_get_code() != CPL_ERROR_NONE){
				esotk_masterfringe_clean_all_recipe_parameters(
					frameset_fringe, NULL, frameset_fringe_bpm,
					frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
					frame_masterfringe_data, frame_masterfringe_error, frame_masterfringe_bpm,
					fringe_mask, NULL,
					himg_masterfringe, hlist_fringe, clist_obj, NULL, NULL, qctable);
				return cpl_error_get_code();
			}

			/* Store the products */
			if (esotk_masterfringe_correct_save(hlist_fringe, parlist, frameset,
					frameset_fringe, qctable, extension, N_ext_start)  != CPL_ERROR_NONE) {
				esotk_masterfringe_clean_all_recipe_parameters(
					frameset_fringe, NULL, frameset_fringe_bpm,
					frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
					frame_masterfringe_data, frame_masterfringe_error, frame_masterfringe_bpm,
					fringe_mask, NULL,
					himg_masterfringe, hlist_fringe, clist_obj, NULL, NULL, qctable);
	            return cpl_error_get_code();
			}

	        /* Clean up */
			if (!single_fringe) {
				cpl_mask_delete(fringe_mask);
				fringe_mask = NULL;
			}
			esotk_masterfringe_clean_loop_recipe_parameters(
				himg_masterfringe, hlist_fringe, clist_obj, NULL, NULL, qctable);
         }
     }

    /* cleanup */
    esotk_clean_recipe_parameters(
		frameset_fringe, NULL, frameset_fringe_bpm,
		frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
		frame_masterfringe_data, frame_masterfringe_error, frame_masterfringe_bpm,
		fringe_mask, NULL);

    return (int)cpl_error_get_code();
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
  * @brief   load masterfringe data into an hdrl_image
  *
  * @param   extension                 input number of the extension
  * @param   frame_masterfringe_data   input masterfringe data
  * @param   frame_masterfringe_bpm    input masterfringe bad pixel mask
  * @param   frame_masterfringe_error  input masterfringe error
  *
  * @return  hdrl_image
 */
/*----------------------------------------------------------------------------*/
hdrl_image * esotk_masterfringe_correct_read(
		const cpl_size extension,
        cpl_frame      *frame_masterfringe_data,
        cpl_frame      *frame_masterfringe_bpm,
        cpl_frame      *frame_masterfringe_error)
{
    hdrl_image *himg_masterfringe = NULL;

    cpl_image *img_data = cpl_image_load(cpl_frame_get_filename(frame_masterfringe_data), HDRL_TYPE_DATA, 0, extension);
    if (img_data == NULL) {
    	/* Error code already set by cpl_image_load. */
        return NULL;
    }

    if (frame_masterfringe_bpm != NULL) {

        cpl_mask *bpm = cpl_mask_load(cpl_frame_get_filename(frame_masterfringe_bpm), 0, extension);
        if (bpm == NULL) {
        	/* Error code already set by cpl_image_load. */
            cpl_image_delete(img_data);
            return NULL;
        }

        cpl_msg_debug(cpl_func, "Number of Bad pixels in masterfringe: %d", (int) cpl_mask_count(bpm));

        cpl_image_reject_from_mask(img_data, bpm);
        cpl_mask_delete(bpm);
    }

    /* if there is the error, ie, frame_masterfringe_error != NULL */
    if (frame_masterfringe_error != NULL) {

        cpl_image *img_error = cpl_image_load(cpl_frame_get_filename(frame_masterfringe_error), HDRL_TYPE_ERROR, 0, extension);
        if (img_error == NULL) {
            cpl_image_delete(img_data);
            return NULL;  /* Error code already set by cpl_image_load. */
        }

        himg_masterfringe = hdrl_image_create(img_data, img_error);

        cpl_image_delete(img_error);

    } else {

        himg_masterfringe = hdrl_image_create(img_data, NULL);
    }

    cpl_image_delete(img_data);

    return himg_masterfringe;
}

/*----------------------------------------------------------------------------*/
/**
  * @brief   save master fringe combination products
  *
  * @param   hlist_fringe     input hdrl_imagelist of fringes
  * @param   parlist          input recipe parameters
  * @param   frameset         input set of frames
  * @param   frameset_inherit input set of frames from where to inherit header
  * @param   qctable          input table with QC parameters, namely Amplitude
  * @param   idx              input number of the current extension
  * @param   N_ext_start      input presence of multiple extensions
  *
  * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_masterfringe_correct_save(
        const hdrl_imagelist    *hlist_fringe,
        const cpl_parameterlist *parlist,
        cpl_frameset            *frameset,
        const cpl_frameset      *frameset_inherit,
        const cpl_table         *qctable,
        const cpl_size          idx,
        const cpl_size          N_ext_start)
{
    const char *type = N_ext_start == 0 ? "start_0" : "start_1";

    if (idx > N_ext_start) {
        type = "append";
    }

    const cpl_size N_img = hdrl_imagelist_get_size(hlist_fringe);
    for (cpl_size frame = 0; frame < N_img; frame++) {

         /* Only pass the relevant QC parameters */
        cpl_table * qctable_local = NULL;
        if (qctable != NULL) {
        	/* in case of rescale parameter set to FALSE */
            qctable_local = cpl_table_extract(qctable, frame, 1);
        } else {
        	cpl_msg_debug(cpl_func, "'qctable' is NULL");
        }

        hdrl_image *hout     = hdrl_imagelist_get(hlist_fringe, frame);
        cpl_image  *out_ima  = hdrl_image_get_image(hout);
        cpl_image  *out_err  = hdrl_image_get_error(hout);

        cpl_mask   *out_bpm  = cpl_image_get_bpm(out_ima);
        cpl_image  *iout_bpm = cpl_image_new_from_mask(out_bpm);

        char *filestr_ima = cpl_sprintf("esotk_fringecorrected_%04d.fits",     (int)frame);
        char *filestr_err = cpl_sprintf("esotk_fringecorrected_%04d_err.fits", (int)frame);
        char *filestr_bpm = cpl_sprintf("esotk_fringecorrected_%04d_bpm.fits", (int)frame);

        const cpl_frame *inherit = cpl_frameset_get_position_const(frameset_inherit, frame);

        esotk_save_image(ESOTK_FRINGE_CORRECTED, RECIPE_NAME,
                         filestr_ima, CPL_TYPE_DOUBLE,
                         out_ima, qctable_local, parlist,
                         frameset, type, inherit);

        esotk_save_image(ESOTK_FRINGE_CORRECTED_ERROR, RECIPE_NAME,
                         filestr_err, CPL_TYPE_DOUBLE,
                         out_err, qctable_local, parlist,
                         frameset, type, inherit);

        esotk_save_image(ESOTK_FRINGE_CORRECTED_BPM, RECIPE_NAME,
                         filestr_bpm, CPL_TYPE_UCHAR,
                         iout_bpm, qctable_local, parlist, frameset, type, inherit);

        cpl_table_delete(qctable_local);

        cpl_image_delete(iout_bpm);

        cpl_free(filestr_ima);
        cpl_free(filestr_err);
        cpl_free(filestr_bpm);
    }

    return cpl_error_get_code();
}
