/*
 * This file is part of the ESO Toolkit
 * Copyright (C) 2016 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                    Includes
 -----------------------------------------------------------------------------*/

#include "esotk_utils.h"
#include "esotk_dfs.h"
#include "esotk_masterfringe_utils.h"

#define HDRL_USE_EXPERIMENTAL
#include "hdrl.h"

#include <cpl.h>

/*----------------------------------------------------------------------------*/
/**
 * @defgroup esotk_masterfringe_create  master fringe creation
 * @par Synopsis: TBD
 * @par Input frames: TBD
 * @par Output frames: TBD
 * @code
 * @endcode
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

/* save image of master fringe */
static cpl_error_code esotk_masterfringe_save(
        const hdrl_image        *master_fringe,
        const cpl_image         *contribution,
        const cpl_parameterlist *parlist,
        cpl_frameset            *frameset,
        const cpl_table         *qctable,
        const cpl_size          idx,
        const cpl_size          N_ext_start);

/*-----------------------------------------------------------------------------
                            Static global variables
 -----------------------------------------------------------------------------*/

#define RECIPE_NAME "esotk_masterfringe_create"

static char esotk_masterfringe_create_description[] =
"                                                                           \n"
"This recipe computes a master-fringe image from a set of dithered          \n"
"input images.                                                              \n"
"                                                                           \n"
"Input files:                                                               \n"
"                                                                           \n"
"  DO category:                      Explanation:           Required:       \n"
"  RAW                               Data                   Yes             \n"
"  RAW_BPM                           Bad pixel masks        NO              \n"
"  RAW_ERROR                         Error image            NO              \n"
"  RAW_VARIANCE                      Variance image         NO              \n"
"  OBJ_MASK                          Object masks           NO              \n"
"  FRINGE_MASK                       Fringe mask            NO              \n"
"                                                                           \n"
"Output files:                                                              \n"
"                                                                           \n"
"  DO category:                      Explanation:                           \n"
"  MASTER_FRINGE                     Master-fringe                          \n"
"  MASTER_FRINGE_ERROR               Error of master-fringe                 \n"
"  MASTER_FRINGE_CONTRIBUTION        Contribution map of master-fringe      \n"
"  MASTER_FRINGE_BPM                 Bad pixel mask of master-fringe        \n"
"                                                                           \n"
"                                                                           \n"
"                                                                           \n"
"                                                                           \n"              
"The algorithm consists of two main steps: i) Using a Gaussian mixture      \n"
"model the recipe estimates the background (a scalar) and the fringe        \n"
"scaling-factor (amplitude - a scalar) of each image tagged as RAW. The     \n"
"background is then subtracted and the amplitude used to normalize the      \n"
"background-subtracted image (multiplicatively) to the same scale. ii)      \n"
"The resulting images are then collapsed into a master-fringe               \n"
"image. There are different collapsing methods available and they are       \n"
"controlled by the -collapse.XXX recipe parameters. In other words: the     \n"
"amplitude of the fringes is computed for each input image (RAW) and        \n"
"used to re-scale the image before stacking.                                \n"
"                                                                           \n"
"Different masks controlling the algorithm in different stages can be       \n"
"passed to the recipe:                                                      \n"
"                                                                           \n"
"A) RAW_BPM: One for each RAW image. Bad pixel mask of each single          \n"
"input image. Each step in the algorithm takes this mask into account.      \n"
"                                                                           \n"
"B) OBJ_MASK: One for each RAW image. Images flagging the objects in        \n"
"the input image. Each step in the algorithm takes this mask into           \n"
"account.                                                                   \n"
"                                                                           \n"
"C) FRINGE_MASK: Single mask. This mask is only taken into account in       \n"
"step i), i.e. when computing the image background and fringe amplitude     \n"
"for each RAW image. Step ii), i.e. the collapsing algorithm ignores        \n"
"this mask. Please note that if the fringe mask has mask-data only in       \n"
"the primary extension or only in the first extension but the raw           \n"
"frames are multi-extension files, the mask-data is applied to all          \n"
"extension. This allows the user to e.g. exclude the borders of all         \n"
"extensions in step i) without the need to create a multi-extension         \n"
"mask-file.                                                                 \n"
"                                                                           \n"
"Please note that part of the code is paralelised. In order to optimise     \n"
"use of computing resources you should set the environment variable         \n"
"OMP_NUM_THREADS to a proper value, like (for bash), to use 4 cores         \n"
"export OMP_NUM_THREADS=4                                                   \n"
"                                                                           \n";

/* Standard CPL recipe definition */
cpl_recipe_define( esotk_masterfringe_create,
                   ESOTK_BINARY_VERSION,
                   "HDRL Group",
                   PACKAGE_BUGREPORT,
                   "2016",
                   "Creates a Master-Fringe",
                   esotk_masterfringe_create_description);

/*----------------------------------------------------------------------------*/
/**
 * @brief Function needed by cpl_recipe_define to fill the input parameters
 *
 * @param  self   parameterlist where you need put parameters
 *
 * @return (int)cpl_error_code
 *
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_masterfringe_create_fill_parameterlist(cpl_parameterlist * self)
{
    /* Master-fringe related parameters */
    hdrl_parameter *sigclip_def = hdrl_collapse_sigclip_parameter_create(3., 3., 5);
    hdrl_parameter *minmax_def  = hdrl_collapse_minmax_parameter_create(1., 1.);
    hdrl_parameter * mode_def =
              hdrl_collapse_mode_parameter_create(10., 1., 0., HDRL_MODE_MEDIAN, 0);

    cpl_parameterlist *pfringecollapse = hdrl_collapse_parameter_create_parlist(
       RECIPE_NAME, "collapse", "MEDIAN", sigclip_def, minmax_def, mode_def);
    hdrl_parameter_delete(sigclip_def);
    hdrl_parameter_delete(minmax_def);
    hdrl_parameter_delete(mode_def);
    
    /* Duplicate in the input/output parameterlist the parameters in the catalogue */
    for (cpl_parameter *p = cpl_parameterlist_get_first(pfringecollapse); p != NULL;
    	p = cpl_parameterlist_get_next(pfringecollapse)) {
        cpl_parameterlist_append(self, cpl_parameter_duplicate(p));
    }
    cpl_parameterlist_delete(pfringecollapse);

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
 * @brief   perform master fringe combination
 *
 * @param   frameset   input set of frames
 * @param   parlist    input recipe parameters
 *
 * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
static int esotk_masterfringe_create(
    cpl_frameset            *frameset,
    const cpl_parameterlist *parlist)
{
    if (esotk_check_and_set_groups(frameset) != CPL_ERROR_NONE) {
    	return cpl_error_get_code();
    }

    cpl_msg_info(cpl_func, "Classifying input frames:");

    cpl_frameset *frameset_fringe       = cpl_frameset_new();
    cpl_frameset *frameset_fringe_bpm   = cpl_frameset_new();
    cpl_frameset *frameset_fringe_obj   = cpl_frameset_new();
    cpl_frameset *frameset_fringe_error = cpl_frameset_new();
    cpl_frameset *frameset_fringe_mask  = cpl_frameset_new();

    cpl_size N_fs = cpl_frameset_get_size(frameset);
    for (cpl_size i = 0; i < N_fs; i++) {

        cpl_frame *cur_frame = cpl_frameset_get_position(frameset, i);

        if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW)) {
            cpl_frameset_insert(frameset_fringe,       cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW_BPM)) {
            cpl_frameset_insert(frameset_fringe_bpm,   cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW_ERROR)) {
            cpl_frameset_insert(frameset_fringe_error, cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_RAW_VARIANCE)) {
        	/* Insert variance maps into the error frameset and convert them
        	 *  into a proper error map when reading - see below */
        	cpl_frameset_insert(frameset_fringe_error, cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_OBJ_MASK)) {
            cpl_frameset_insert(frameset_fringe_obj,   cpl_frame_duplicate(cur_frame));
        } else if (!strcmp(cpl_frame_get_tag(cur_frame), ESOTK_FRINGE_MASK)) {
            cpl_frameset_insert(frameset_fringe_mask , cpl_frame_duplicate(cur_frame));
        }
    }

    hdrl_parameter *collapse_params = hdrl_collapse_parameter_parse_parlist(parlist, RECIPE_NAME".collapse");

    /* Check integrity */
    if (esotk_integrity_checks(
		 frameset_fringe, NULL, frameset_fringe_bpm,
		 frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
		 NULL, NULL, NULL, NULL, collapse_params) != CPL_ERROR_NONE) {
		return (int)cpl_error_get_code();
	}

    /* load fringe mask */
    cpl_mask *fringe_mask = esotk_masterfringe_load_fringe_mask(frameset_fringe_mask);
    cpl_boolean single_fringe = fringe_mask != NULL ? CPL_TRUE : CPL_FALSE;


    cpl_msg_info(cpl_func, "Starting masterfringe computation:");

    /* If there are multi extension frames the primary extension is empty */
    const cpl_size N_mask = cpl_frameset_get_size(frameset_fringe_mask );
    const cpl_size N_ext  = cpl_frame_get_nextensions(cpl_frameset_get_position(frameset_fringe, 0) );
    const cpl_size N_ext_start = N_ext > 0 ? 1 : 0;
    for (cpl_size extension = N_ext_start; extension <= N_ext; extension++) {

        cpl_msg_info(cpl_func, "Working on extension %"CPL_SIZE_FORMAT, extension);

        /* no single global fringe, load fringe mask of extension */
        if (N_mask == 1 && fringe_mask == NULL) {
            cpl_frame  *f_mask = cpl_frameset_get_position(frameset_fringe_mask, 0);
            const char *fn     = cpl_frame_get_filename(f_mask);
            fringe_mask        = cpl_mask_load(fn, 0, extension);
        }

		/* Loop over the frames */
        hdrl_imagelist *hlist_fringe = hdrl_imagelist_new();
        cpl_imagelist  *clist_obj    = cpl_imagelist_new();
        if (esotk_masterfringe_loop_img_frames(extension, hlist_fringe, clist_obj,
        		frameset_fringe, frameset_fringe_bpm, frameset_fringe_obj,
				frameset_fringe_error) != CPL_ERROR_NONE) {
        	esotk_masterfringe_clean_all_recipe_parameters(
				frameset_fringe, NULL, frameset_fringe_bpm,
				frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
				NULL, NULL, NULL, fringe_mask, collapse_params,
				NULL, hlist_fringe, clist_obj, NULL, NULL, NULL);
            return (int)cpl_error_get_code();
        }

        /* compute master fringe */
        hdrl_image *master;
        cpl_image  *contrib_map;
        cpl_table  *qctable;
        hdrl_fringe_compute(hlist_fringe, clist_obj, fringe_mask,
        	collapse_params, &master, &contrib_map, &qctable);

        /* Check computing */
		if (cpl_error_get_code() != CPL_ERROR_NONE){
        	esotk_masterfringe_clean_all_recipe_parameters(
				frameset_fringe, NULL, frameset_fringe_bpm,
				frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
				NULL, NULL, NULL, fringe_mask, collapse_params,
			    NULL, hlist_fringe, clist_obj, NULL, NULL, NULL);
            return cpl_error_get_code();
		}

        /* Store master fringe */
        if (esotk_masterfringe_save(master, contrib_map, parlist, frameset,
        		qctable, extension, N_ext_start)  != CPL_ERROR_NONE) {
        	esotk_masterfringe_clean_all_recipe_parameters(
				frameset_fringe, NULL, frameset_fringe_bpm,
				frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
				NULL, NULL, NULL, fringe_mask, collapse_params,
				NULL, hlist_fringe, clist_obj, NULL, NULL, NULL);
            return cpl_error_get_code();
        }

        /* Clean up */
        if (!single_fringe) {
    		/* Free the fringe mask for the next extension */
            cpl_mask_delete(fringe_mask);
            fringe_mask = NULL;
        }
		esotk_masterfringe_clean_loop_recipe_parameters(
			NULL, hlist_fringe, clist_obj, master, contrib_map, qctable);
    }

    /* cleanup */
    esotk_clean_recipe_parameters(
		frameset_fringe, NULL, frameset_fringe_bpm,
		frameset_fringe_obj, frameset_fringe_error, frameset_fringe_mask,
		NULL, NULL, NULL, fringe_mask, collapse_params);

    return (int)cpl_error_get_code();
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
 * @brief   save master fringe combination products
 *
 * @param   master_fringe    input image of master fringe
 * @param   contribution     input contribution
 * @param   parlist          input set of frames
 * @param   frameset         input set of frames from where to inherit header
 * @param   qctable          input table with QC parameters, namely Amplitude
 * @param   idx              input number of the current extension
 * @param   N_ext_start      input presence of multiple extensions
 *
 * @return  cpl_error_code
 */
/*----------------------------------------------------------------------------*/
static cpl_error_code esotk_masterfringe_save(
        const hdrl_image        *master_fringe,
        const cpl_image         *contribution,
        const cpl_parameterlist *parlist,
        cpl_frameset            *frameset,
        const cpl_table         *qctable,
        const cpl_size          idx,
        const cpl_size          N_ext_start)
{
    const char *type = N_ext_start == 0 ? "start_0" : "start_1";

    if (idx > N_ext_start) {
        type = "append";
    }

    esotk_save_image(ESOTK_MASTER_FRINGE, RECIPE_NAME,
                     "esotk_masterfringe.fits", CPL_TYPE_DOUBLE,
                     hdrl_image_get_image_const(master_fringe), qctable, parlist,
                     frameset, type, NULL);

    esotk_save_image(ESOTK_MASTER_FRINGE_ERROR, RECIPE_NAME,
                     "esotk_masterfringe_error.fits", CPL_TYPE_DOUBLE,
                     hdrl_image_get_error_const(master_fringe), qctable, parlist,
                     frameset, type, NULL);

    esotk_save_image(ESOTK_MASTER_FRINGE_CONTRIBUTION, RECIPE_NAME,
                     "esotk_masterfringe_contribution.fits", CPL_TYPE_INT,
                     contribution, qctable, parlist, frameset, type, NULL);

    cpl_mask  *bpm  = cpl_mask_threshold_image_create(contribution, -0.5, 0.5);
    cpl_image *ibpm = cpl_image_new_from_mask(bpm);

    esotk_save_image(ESOTK_MASTER_FRINGE_BPM, RECIPE_NAME,
                     "esotk_masterfringe_bpm.fits", CPL_TYPE_UCHAR,
                     ibpm, qctable, parlist, frameset, type, NULL);

    cpl_mask_delete(bpm);
    cpl_image_delete(ibpm);

    return cpl_error_get_code();
}
